import os
from pathlib import Path
from shutil import copyfile, rmtree
from typing import Any, Dict

import numpy as np
import pynwb
import pytest
from pytest_mock import MockerFixture
import zarr

from .fixtures import SampleDandiset
from .test_helpers import assert_dirtrees_eq
from ..consts import ZARR_MIME_TYPE, dandiset_metadata_file
from ..dandiapi import AssetType, RemoteBlobAsset, RemoteZarrAsset
from ..dandiset import APIDandiset
from ..download import download
from ..exceptions import NotFoundError
from ..files import LocalFileAsset
from ..pynwb_utils import make_nwb_file
from ..utils import list_paths, yaml_dump


def test_upload_download(
    new_dandiset: SampleDandiset, organized_nwb_dir: str, tmp_path: Path
) -> None:
    d = new_dandiset.dandiset
    dspath = new_dandiset.dspath
    (nwb_file,) = [
        p for p in list_paths(organized_nwb_dir) if p.name != dandiset_metadata_file
    ]
    assert nwb_file.suffix == ".nwb"
    parent, name = nwb_file.relative_to(organized_nwb_dir).parts
    (dspath / parent).mkdir()
    copyfile(nwb_file, dspath / parent / name)
    new_dandiset.upload()
    download(d.version_api_url, tmp_path)
    assert list_paths(tmp_path) == [
        tmp_path / d.identifier / dandiset_metadata_file,
        tmp_path / d.identifier / parent / name,
    ]


def test_upload_dandiset_metadata(new_dandiset: SampleDandiset) -> None:
    # For now let's "manually" populate dandiset.yaml in that downloaded location
    # which is missing due to https://github.com/dandi/dandi-api/issues/63
    d = new_dandiset.dandiset
    dspath = new_dandiset.dspath
    ds_orig = APIDandiset(dspath)
    ds_metadata = ds_orig.metadata
    assert ds_metadata is not None
    ds_metadata["description"] = "very long"
    ds_metadata["name"] = "shorty"
    (dspath / dandiset_metadata_file).write_text(yaml_dump(ds_metadata))
    new_dandiset.upload(
        paths=[dspath / dandiset_metadata_file], upload_dandiset_metadata=True
    )
    d.refresh()
    assert d.version.name == "shorty"


def test_upload_extant_existing(
    mocker: MockerFixture, text_dandiset: SampleDandiset
) -> None:
    iter_upload_spy = mocker.spy(LocalFileAsset, "iter_upload")
    with pytest.raises(FileExistsError):
        text_dandiset.upload(existing="error")
    iter_upload_spy.assert_not_called()


def test_upload_extant_skip(
    mocker: MockerFixture, text_dandiset: SampleDandiset
) -> None:
    iter_upload_spy = mocker.spy(LocalFileAsset, "iter_upload")
    text_dandiset.upload(existing="skip")
    iter_upload_spy.assert_not_called()


@pytest.mark.parametrize("existing", ["overwrite", "refresh"])
def test_upload_extant_eq_overwrite(
    existing: str, mocker: MockerFixture, text_dandiset: SampleDandiset
) -> None:
    iter_upload_spy = mocker.spy(LocalFileAsset, "iter_upload")
    text_dandiset.upload(existing=existing)
    iter_upload_spy.assert_not_called()


@pytest.mark.parametrize("existing", ["overwrite", "refresh"])
def test_upload_extant_neq_overwrite(
    existing: str, mocker: MockerFixture, text_dandiset: SampleDandiset, tmp_path: Path
) -> None:
    (text_dandiset.dspath / "file.txt").write_text("This is different text.\n")
    iter_upload_spy = mocker.spy(LocalFileAsset, "iter_upload")
    text_dandiset.upload(existing=existing)
    iter_upload_spy.assert_called()
    download(text_dandiset.dandiset.version_api_url, tmp_path)
    assert (
        tmp_path / text_dandiset.dandiset_id / "file.txt"
    ).read_text() == "This is different text.\n"


def test_upload_extant_old_refresh(
    mocker: MockerFixture, text_dandiset: SampleDandiset
) -> None:
    (text_dandiset.dspath / "file.txt").write_text("This is different text.\n")
    os.utime(text_dandiset.dspath / "file.txt", times=(0, 0))
    iter_upload_spy = mocker.spy(LocalFileAsset, "iter_upload")
    text_dandiset.upload(existing="refresh")
    iter_upload_spy.assert_not_called()


def test_upload_extant_force(
    mocker: MockerFixture, text_dandiset: SampleDandiset
) -> None:
    iter_upload_spy = mocker.spy(LocalFileAsset, "iter_upload")
    text_dandiset.upload(existing="force")
    iter_upload_spy.assert_called()


def test_upload_extant_bad_existing(
    mocker: MockerFixture, text_dandiset: SampleDandiset
) -> None:
    iter_upload_spy = mocker.spy(LocalFileAsset, "iter_upload")
    text_dandiset.upload(existing="foobar")
    iter_upload_spy.assert_not_called()


@pytest.mark.parametrize(
    "contents",
    [
        pytest.param(
            b"",
            marks=pytest.mark.xfail(
                reason="https://github.com/dandi/dandi-api/issues/168"
            ),
        ),
        b"x",
    ],
)
def test_upload_download_small_file(
    contents: bytes, new_dandiset: SampleDandiset, tmp_path: Path
) -> None:
    d = new_dandiset.dandiset
    dandiset_id = d.identifier
    dspath = new_dandiset.dspath
    (dspath / "file.txt").write_bytes(contents)
    new_dandiset.upload(allow_any_path=True)
    download(d.version_api_url, tmp_path)
    assert list_paths(tmp_path) == [
        tmp_path / dandiset_id / dandiset_metadata_file,
        tmp_path / dandiset_id / "file.txt",
    ]
    assert (tmp_path / dandiset_id / "file.txt").read_bytes() == contents


@pytest.mark.parametrize("confirm", [True, False])
def test_upload_sync(
    confirm: bool, mocker: MockerFixture, text_dandiset: SampleDandiset
) -> None:
    (text_dandiset.dspath / "file.txt").unlink()
    confirm_mock = mocker.patch("click.confirm", return_value=confirm)
    text_dandiset.upload(sync=True)
    confirm_mock.assert_called_with("Delete 1 asset on server?")
    if confirm:
        with pytest.raises(NotFoundError):
            text_dandiset.dandiset.get_asset_by_path("file.txt")
    else:
        text_dandiset.dandiset.get_asset_by_path("file.txt")


def test_upload_sync_folder(
    mocker: MockerFixture, text_dandiset: SampleDandiset
) -> None:
    (text_dandiset.dspath / "file.txt").unlink()
    (text_dandiset.dspath / "subdir2" / "banana.txt").unlink()
    confirm_mock = mocker.patch("click.confirm", return_value=True)
    text_dandiset.upload(paths=[text_dandiset.dspath / "subdir2"], sync=True)
    confirm_mock.assert_called_with("Delete 1 asset on server?")
    text_dandiset.dandiset.get_asset_by_path("file.txt")
    with pytest.raises(NotFoundError):
        text_dandiset.dandiset.get_asset_by_path("subdir2/banana.txt")


def test_upload_sync_zarr(mocker, zarr_dandiset):
    rmtree(zarr_dandiset.dspath / "sample.zarr")
    zarr.save(zarr_dandiset.dspath / "identity.zarr", np.eye(5))
    confirm_mock = mocker.patch("click.confirm", return_value=True)
    zarr_dandiset.upload(sync=True)
    confirm_mock.assert_called_with("Delete 1 asset on server?")
    zarr_dandiset.dandiset.get_asset_by_path("identity.zarr")
    with pytest.raises(NotFoundError):
        zarr_dandiset.dandiset.get_asset_by_path("sample.zarr")


def test_upload_invalid_metadata(
    new_dandiset: SampleDandiset, simple1_nwb_metadata: Dict[str, Any]
) -> None:
    make_nwb_file(
        new_dandiset.dspath / "broken.nwb",
        subject=pynwb.file.Subject(
            subject_id="mouse001",
            age="XLII anni",
            sex="yes",
            species="unicorn",
        ),
        **simple1_nwb_metadata,
    )
    new_dandiset.upload()
    with pytest.raises(NotFoundError):
        new_dandiset.dandiset.get_asset_by_path("broken.nwb")


def test_upload_zarr(new_dandiset: SampleDandiset) -> None:
    zarr.save(
        new_dandiset.dspath / "sample.zarr", np.arange(1000), np.arange(1000, 0, -1)
    )
    new_dandiset.upload()
    (asset,) = new_dandiset.dandiset.get_assets()
    assert isinstance(asset, RemoteZarrAsset)
    assert asset.asset_type is AssetType.ZARR
    assert asset.path == "sample.zarr"


def test_upload_different_zarr(tmp_path: Path, zarr_dandiset: SampleDandiset) -> None:
    asset = zarr_dandiset.dandiset.get_asset_by_path("sample.zarr")
    assert isinstance(asset, RemoteZarrAsset)
    zarr_id = asset.zarr
    rmtree(zarr_dandiset.dspath / "sample.zarr")
    zarr.save(zarr_dandiset.dspath / "sample.zarr", np.eye(5))
    zarr_dandiset.upload()
    asset = zarr_dandiset.dandiset.get_asset_by_path("sample.zarr")
    assert isinstance(asset, RemoteZarrAsset)
    assert asset.zarr == zarr_id
    download(zarr_dandiset.dandiset.version_api_url, tmp_path)
    assert_dirtrees_eq(
        zarr_dandiset.dspath / "sample.zarr",
        tmp_path / zarr_dandiset.dandiset_id / "sample.zarr",
    )


def test_upload_different_zarr_entry_conflicts(
    tmp_path: Path, new_dandiset: SampleDandiset
) -> None:
    zf = new_dandiset.dspath / "sample.zarr"
    zf.mkdir()
    (zf / "unchanged.txt").write_text("This is will not change.\n")
    (zf / "changed-contents.txt").write_text("This is text version #1.\n")
    (zf / "changed-size.txt").write_text("This is a test.\n")
    (zf / "changed-type").mkdir()
    (zf / "changed-type" / "file.txt").write_text("This is test text.\n")
    new_dandiset.upload(validation="skip")
    rmtree(zf)
    zf.mkdir()
    (zf / "unchanged.txt").write_text("This is will not change.\n")
    (zf / "changed-contents.txt").write_text("This is text version #2.\n")
    (zf / "changed-size.txt").write_text("This is a test of the upload code.\n")
    (zf / "changed-type").write_text("This is now a file.\n")
    new_dandiset.upload(validation="skip")
    download(new_dandiset.dandiset.version_api_url, tmp_path)
    assert_dirtrees_eq(zf, tmp_path / new_dandiset.dandiset_id / "sample.zarr")


def test_upload_different_zarr_file_to_parent_dir(
    tmp_path: Path, new_dandiset: SampleDandiset
) -> None:
    zf = new_dandiset.dspath / "sample.zarr"
    zf.mkdir()
    (zf / "foo").write_text("This is a file.\n")
    new_dandiset.upload(validation="skip")
    rmtree(zf)
    zf.mkdir()
    (zf / "foo").mkdir()
    (zf / "foo" / "bar").write_text("This is under what used to be a file.\n")
    new_dandiset.upload(validation="skip")
    download(new_dandiset.dandiset.version_api_url, tmp_path)
    assert_dirtrees_eq(zf, tmp_path / new_dandiset.dandiset_id / "sample.zarr")


def test_upload_nonzarr_to_zarr_path(
    tmp_path: Path, zarr_dandiset: SampleDandiset
) -> None:
    rmtree(zarr_dandiset.dspath / "sample.zarr")
    (zarr_dandiset.dspath / "sample.zarr").write_text("This is not a Zarr.\n")
    zarr_dandiset.upload(allow_any_path=True)
    (asset,) = zarr_dandiset.dandiset.get_assets()
    assert isinstance(asset, RemoteBlobAsset)
    assert asset.asset_type is AssetType.BLOB
    assert asset.path == "sample.zarr"
    assert asset.get_raw_metadata()["encodingFormat"] == "application/octet-stream"
    download(zarr_dandiset.dandiset.version_api_url, tmp_path)
    assert (
        tmp_path / zarr_dandiset.dandiset_id / "sample.zarr"
    ).read_text() == "This is not a Zarr.\n"


def test_upload_zarr_to_nonzarr_path(
    new_dandiset: SampleDandiset, tmp_path: Path
) -> None:
    d = new_dandiset.dandiset
    dspath = new_dandiset.dspath
    (dspath / "sample.zarr").write_text("This is not a Zarr.\n")
    new_dandiset.upload(allow_any_path=True)

    (asset,) = d.get_assets()
    assert isinstance(asset, RemoteBlobAsset)
    assert asset.asset_type is AssetType.BLOB
    assert asset.path == "sample.zarr"
    assert asset.get_raw_metadata()["encodingFormat"] == "application/octet-stream"

    (dspath / "sample.zarr").unlink()
    zarr.save(dspath / "sample.zarr", np.arange(1000), np.arange(1000, 0, -1))
    new_dandiset.upload(allow_any_path=True)

    (asset,) = d.get_assets()
    assert isinstance(asset, RemoteZarrAsset)
    assert asset.asset_type is AssetType.ZARR
    assert asset.path == "sample.zarr"
    assert asset.get_raw_metadata()["encodingFormat"] == ZARR_MIME_TYPE

    download(d.version_api_url, tmp_path)
    assert_dirtrees_eq(dspath / "sample.zarr", tmp_path / d.identifier / "sample.zarr")


def test_upload_zarr_with_empty_dir(new_dandiset: SampleDandiset) -> None:
    zarr.save(
        new_dandiset.dspath / "sample.zarr", np.arange(1000), np.arange(1000, 0, -1)
    )
    (new_dandiset.dspath / "sample.zarr" / "empty").mkdir()
    new_dandiset.upload()
    (asset,) = new_dandiset.dandiset.get_assets()
    assert isinstance(asset, RemoteZarrAsset)
    assert asset.asset_type is AssetType.ZARR
    assert asset.path == "sample.zarr"
    assert not (asset.filetree / "empty").exists()



import click

from .base import dandiset_path_option, devel_debug_option, map_to_click_exceptions
from ..consts import file_operation_modes


@click.command()
@dandiset_path_option(
    help="The root directory of the Dandiset to organize files under. "
    "If not specified, the Dandiset under the current directory is assumed. "
    "For 'simulate' mode, the target Dandiset/directory must not exist.",
    type=click.Path(dir_okay=True, file_okay=False),
)
@click.option(
    "--invalid",
    help="What to do if files without sufficient metadata are encountered.",
    type=click.Choice(["fail", "warn"]),
    default="fail",
    show_default=True,
)
@click.option(
    "-f",
    "--files-mode",
    help="If 'dry' - no action is performed, suggested renames are printed. "
    "If 'simulate' - hierarchy of empty files at --dandiset-path is created. "
    "Note that previous layout should be removed prior to this operation.  "
    "If 'auto' - whichever of symlink, hardlink, copy is allowed by system. "
    "The other modes (copy, move, symlink, hardlink) define how data files "
    "should be made available.",
    type=click.Choice(file_operation_modes),
    default="auto",
    show_default=True,
)
@click.option(
    "--update-external-file-paths",
    is_flag=True,
    default=False,
    help="Rewrite the 'external_file' arguments of ImageSeries in NWB files. "
    "The new values will correspond to the new locations of the video files "
    "after being organized. "
    "This option requires --files-mode to be 'copy' or 'move'",
)
@click.option(
    "--media-files-mode",
    type=click.Choice(["copy", "move", "symlink", "hardlink"]),
    default=None,
    help="How to relocate video files referenced by NWB files",
)
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
@devel_debug_option()
@map_to_click_exceptions
def organize(
    paths,
    dandiset_path=None,
    invalid="fail",
    files_mode="auto",
    devel_debug=False,
    update_external_file_paths=False,
    media_files_mode=None,
):
    """(Re)organize files according to the metadata.

    The purpose of this command is to take advantage of metadata contained in
    .nwb files to provide datasets with consistently-named files whose names
    reflect the data they contain.

    .nwb files are organized into a hierarchy of subfolders, one per "subject",
    e.g., `sub-0001` if an .nwb file contained a Subject group with
    `subject_id=0001`.  Each file in a subject-specific subfolder follows the
    pattern:

        sub-<subject_id>[_key-<value>][_mod1+mod2+...].nwb

    where the following keys are considered if present in the data:

    \b
        ses -- session_id
        tis -- tissue_sample_id
        slice -- slice_id
        cell -- cell_id

    and `modX` are "modalities" as identified based on detected neural data
    types (such as "ecephys", "icephys") per extensions found in nwb-schema
    definitions:
    https://github.com/NeurodataWithoutBorders/nwb-schema/tree/dev/core

    In addition, an "obj" key with a value corresponding to the crc32 checksum
    of "object_id" is added if the aforementioned keys and the list of
    modalities are not sufficient to disambiguate different files.

    You can visit https://dandiarchive.org for a growing collection of
    (re)organized dandisets.
    """
    from ..organize import organize

    organize(
        paths,
        dandiset_path=dandiset_path,
        invalid=invalid,
        files_mode=files_mode,
        devel_debug=devel_debug,
        update_external_file_paths=update_external_file_paths,
        media_files_mode=media_files_mode,
    )



import os

from click.testing import CliRunner


def test_validate_bids_error(bids_examples):
    from ..cmd_validate import validate_bids

    expected_error = (
        "Summary: 2 filename patterns required by BIDS could not be found "
        "and 1 filename did not match any pattern known to BIDS.\n"
    )
    broken_dataset = os.path.join(bids_examples, "invalid_pet001")
    r = CliRunner().invoke(validate_bids, [broken_dataset])
    # Does it break?
    assert r.exit_code == 1

    # Does it report the issue correctly?
    assert r.output == expected_error



"""
THIS MODULE IS BUNDLED FROM THE bids-specification PACKAGE.
Utility functions for the bids-specification schema.
"""
import logging
import os.path as op
from pprint import pprint

import numpy as np


def get_schema_path():
    """Get the path to the schema folder.

    Returns
    -------
    str
        Absolute path to the folder containing schema-related files.
    """
    return op.abspath(op.join(op.dirname(__file__), "data", "schema"))


def combine_extensions(lst):
    """Combine extensions with their compressed versions in a list.

    This is a basic solution to combining extensions with their
    compressed versions in a list. Something more robust could
    be written in the future.

    Parameters
    ----------
    lst : list of str
    """
    new_lst = []
    # First, sort by length
    lst = sorted(lst, key=len)
    for item in lst:
        temp_lst = new_lst[:]

        item_found = False
        for j, new_item in enumerate(temp_lst):
            if new_item in item:
                temp_item = new_item + "[" + item.replace(new_item, "", 1) + "]"
                new_lst[j] = temp_item
                item_found = True

        if not item_found:
            new_lst.append(item)

    return new_lst


def get_logger(name=None):
    """Return a logger to use.

    Parameters
    ----------
    name : None or str, optional
        Name of the logger. Defaults to None.

    Returns
    -------
    logging.Logger
        logger object.
    """
    return logging.getLogger("bids-schema" + (".%s" % name if name else ""))


def set_logger_level(lgr, level):
    """Set the logger level.

    Parameters
    ----------
    lgr : logging.Logger
        logger object for which to change the level.
    level : int or str
        The desired level for the logger.
    """
    if isinstance(level, int):
        pass
    elif level.isnumeric():
        level = int(level)
    elif level.isalpha():
        level = getattr(logging, level)
    else:
        lgr.warning("Do not know how to treat loglevel %s" % level)
        return
    lgr.setLevel(level)


def drop_unused_entities(df):
    """Remove columns from a dataframe where all values in the column are NaNs.

    For entity tables, this limits each table to only entities that are used
    within the modality.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame containing entities and datatypes/suffixes.
        Rows are datatype/suffix combinations and columns are entities.

    Returns
    -------
    df : pandas.DataFrame
        DataFrame with columns associated with unused entities removed.
    """
    df = df.replace("", np.nan).dropna(axis=1, how="all").fillna("")
    return df


def flatten_multiindexed_columns(df):
    """Remove multi-indexing of multi-indexed column headers.

    The first layer is the "DataType", while the second layer is the "Format".
    This second layer will become a new row.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame with two header levels: "Datatype" and "Format".

    Returns
    -------
    df : pandas.DataFrame
        DataFrame with the second header level ("Format") converted to a
        normal row.
    """
    # Flatten multi-index
    vals = df.index.tolist()
    df.loc["Format"] = df.columns.get_level_values(1)
    df.columns = df.columns.get_level_values(0)
    df = df.loc[["Format"] + vals]
    df.index.name = "Entity"
    df = df.drop(columns=["DataType"])
    return df


def get_link(string):
    refs = {
        "array": "https://www.w3schools.com/js/js_json_arrays.asp",
        "string": "https://www.w3schools.com/js/js_json_datatypes.asp",
        "number": "https://www.w3schools.com/js/js_json_datatypes.asp",
        "object": "https://www.json.org/json-en.html",
        "integer": "https://www.w3schools.com/js/js_json_datatypes.asp",
        "boolean": "https://www.w3schools.com/js/js_json_datatypes.asp",
    }
    # Allow plurals (e.g., strings -> links to string)
    dtype = string[:-1] if string[-1] == "s" else string
    url = refs.get(dtype)
    if url:
        return f"[{string}]({url})"
    return string


def resolve_metadata_type(definition):
    """Generate string of metadata type from dictionary.

    Parameters
    ----------
    definition : :obj:`dict`
        A schema object definition for a metadata term.

    Returns
    -------
    string : :obj:`str`
        A string describing the valid value types for the metadata term.
    """
    if "type" in definition.keys():
        string = get_link(definition["type"])

        if definition.get("enum") == ["n/a"]:
            # Special string case of n/a
            string = '`"n/a"`'

        elif "type" in definition.get("items", {}):
            # Items within arrays
            string += " of " + get_link(definition["items"]["type"] + "s")

        elif "type" in definition.get("additionalProperties", {}):
            # Values within objects
            string += " of " + get_link(definition["additionalProperties"]["type"] + "s")

    elif "anyOf" in definition:
        # Use dictionary to get unique substrings while preserving insertion order
        substrings = {resolve_metadata_type(subdict): None for subdict in definition["anyOf"]}

        string = " or ".join(substrings)

    else:
        # A hack to deal with $ref in the current schema
        print(f"Type could not be inferred for {definition['name']}")
        pprint(definition)
        string = "unknown"

    return string



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
import logging
import os
from pathlib import Path
import re
import shutil
from subprocess import DEVNULL, check_output, run
import sys
import tempfile
from time import sleep
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterator, List, Optional, Union
from uuid import uuid4

from _pytest.fixtures import FixtureRequest
from click.testing import CliRunner
from dandischema.consts import DANDI_SCHEMA_VERSION
from dateutil.tz import tzlocal, tzutc
import numpy as np
import pynwb
from pynwb import NWBHDF5IO, NWBFile
from pynwb.device import Device
from pynwb.file import Subject
import pynwb.image
from pynwb.ophys import ImageSeries
import pytest
import requests
import zarr

from .skip import skipif
from .. import get_logger
from ..cli.command import organize
from ..consts import DandiInstance, dandiset_metadata_file, known_instances
from ..dandiapi import DandiAPIClient, RemoteDandiset
from ..pynwb_utils import make_nwb_file, metadata_nwb_file_fields
from ..upload import upload

lgr = get_logger()


@pytest.fixture(autouse=True)
def capture_all_logs(caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level(logging.DEBUG, logger="dandi")


# TODO: move into some common fixtures.  We might produce a number of files
#       and also carry some small ones directly in git for regression testing
@pytest.fixture(scope="session")
def simple1_nwb_metadata() -> Dict[str, Any]:
    # very simple assignment with the same values as the key with 1 as suffix
    metadata: Dict[str, Any] = {f: f"{f}1" for f in metadata_nwb_file_fields}
    metadata["identifier"] = uuid4().hex
    # subject_fields

    # tune specific ones:
    # Needs an explicit time zone since otherwise pynwb would add one
    # But then comparison breaks anyways any ways yoh have tried to set it
    # for datetime.now.  Taking example from pynwb tests
    metadata["session_start_time"] = datetime(2017, 4, 15, 12, tzinfo=tzutc())
    metadata["keywords"] = ["keyword1", "keyword 2"]
    # since NWB 2.1.0 some fields values become "arrays" which are
    # then reloaded as tuples, so we force them being tuples here
    # See e.g. https://github.com/NeurodataWithoutBorders/pynwb/issues/1091
    for f in "related_publications", "experimenter":
        metadata[f] = (metadata[f],)
    return metadata


@pytest.fixture(scope="session")
def simple1_nwb(
    simple1_nwb_metadata: Dict[str, Any], tmpdir_factory: pytest.TempdirFactory
) -> str:
    return make_nwb_file(
        str(tmpdir_factory.mktemp("simple1").join("simple1.nwb")),
        **simple1_nwb_metadata,
    )


@pytest.fixture(scope="session")
def simple2_nwb(
    simple1_nwb_metadata: Dict[str, Any], tmpdir_factory: pytest.TempdirFactory
) -> str:
    """With a subject"""
    return make_nwb_file(
        str(tmpdir_factory.mktemp("simple2").join("simple2.nwb")),
        subject=pynwb.file.Subject(
            subject_id="mouse001",
            date_of_birth=datetime(2016, 12, 1, tzinfo=tzutc()),
            sex="M",
            species="mouse",
        ),
        **simple1_nwb_metadata,
    )


@pytest.fixture(scope="session")
def simple3_nwb(
    simple1_nwb_metadata: Dict[str, Any], tmpdir_factory: pytest.TempdirFactory
) -> str:
    """With a subject, but no subject_id."""
    return make_nwb_file(
        str(tmpdir_factory.mktemp("simple2").join("simple2.nwb")),
        subject=pynwb.file.Subject(
            date_of_birth=datetime(2016, 12, 1, tzinfo=tzutc()),
            sex="M",
            species="mouse",
        ),
        **simple1_nwb_metadata,
    )


@pytest.fixture(scope="session")
def organized_nwb_dir(
    simple2_nwb: str, tmp_path_factory: pytest.TempPathFactory
) -> Path:
    tmp_path = tmp_path_factory.mktemp("organized_nwb_dir")
    (tmp_path / dandiset_metadata_file).write_text("{}\n")
    r = CliRunner().invoke(
        organize, ["-f", "copy", "--dandiset-path", str(tmp_path), str(simple2_nwb)]
    )
    assert r.exit_code == 0, r.stdout
    return tmp_path


@pytest.fixture(scope="session")
def organized_nwb_dir2(
    simple1_nwb_metadata: Dict[str, Any],
    simple2_nwb: str,
    tmp_path_factory: pytest.TempPathFactory,
) -> Path:
    tmp_path = tmp_path_factory.mktemp("organized_nwb_dir2")

    # need to copy first and then use -f move since we will create one more
    # file to be "organized"
    shutil.copy(simple2_nwb, tmp_path)
    make_nwb_file(
        str(tmp_path / "simple3.nwb"),
        subject=pynwb.file.Subject(
            subject_id="lizard001",
            date_of_birth=datetime(2016, 12, 1, tzinfo=tzutc()),
            sex="F",
            species="Gekko gecko",
        ),
        **simple1_nwb_metadata,
    )
    (tmp_path / dandiset_metadata_file).write_text("{}\n")
    r = CliRunner().invoke(organize, ["-f", "move", "--dandiset-path", str(tmp_path)])
    assert r.exit_code == 0, r.stdout
    assert sum(p.is_dir() for p in tmp_path.iterdir()) == 2
    return tmp_path


if TYPE_CHECKING:
    if sys.version_info >= (3, 8):
        from typing import Literal
    else:
        from typing_extensions import Literal

    Scope = Union[
        Literal["session"],
        Literal["package"],
        Literal["module"],
        Literal["class"],
        Literal["function"],
    ]


def get_gitrepo_fixture(
    url: str,
    committish: Optional[str] = None,
    scope: Scope = "session",
) -> Callable[[], Iterator[str]]:

    if committish:
        raise NotImplementedError()

    @pytest.fixture(scope=scope)
    def fixture() -> Iterator[str]:
        skipif.no_network()
        skipif.no_git()

        path = tempfile.mktemp()  # not using pytest's tmpdir fixture to not
        # collide in different scopes etc. But we
        # would need to remove it ourselves
        lgr.debug("Cloning %r into %r", url, path)
        try:
            runout = run(["git", "clone", "--depth=1", url, path])
            if runout.returncode:
                raise RuntimeError(f"Failed to clone {url} into {path}")
            yield path
        finally:
            try:
                shutil.rmtree(path)
            except BaseException as exc:
                lgr.warning("Failed to remove %s - using Windows?: %s", path, exc)

    return fixture


nwb_test_data = get_gitrepo_fixture("http://github.com/dandi-datasets/nwb_test_data")
bids_examples = get_gitrepo_fixture("https://github.com/dandi/bids-examples")

LOCAL_DOCKER_DIR = Path(__file__).with_name("data") / "dandiarchive-docker"
LOCAL_DOCKER_ENV = LOCAL_DOCKER_DIR.name


@pytest.fixture(scope="session")
def docker_compose_setup() -> Iterator[Dict[str, str]]:
    skipif.no_network()
    skipif.no_docker_engine()

    # Check that we're running on a Unix-based system (Linux or macOS), as the
    # Docker images don't work on Windows.
    if os.name != "posix":
        pytest.skip("Docker images require Unix host")

    persist = os.environ.get("DANDI_TESTS_PERSIST_DOCKER_COMPOSE")

    create = (
        persist is None
        or run(
            ["docker", "inspect", f"{LOCAL_DOCKER_ENV}_django_1"],
            stdout=DEVNULL,
            stderr=DEVNULL,
        ).returncode
        != 0
    )

    env = {**os.environ, "DJANGO_DANDI_SCHEMA_VERSION": DANDI_SCHEMA_VERSION}
    try:
        if create:
            run(
                ["docker-compose", "run", "--rm", "django", "./manage.py", "migrate"],
                cwd=str(LOCAL_DOCKER_DIR),
                env=env,
                check=True,
            )
            run(
                [
                    "docker-compose",
                    "run",
                    "--rm",
                    "-e",
                    "DJANGO_SUPERUSER_PASSWORD=nsNc48DBiS",
                    "django",
                    "./manage.py",
                    "createsuperuser",
                    "--no-input",
                    "--email",
                    "admin@nil.nil",
                ],
                cwd=str(LOCAL_DOCKER_DIR),
                env=env,
                check=True,
            )

        r = check_output(
            [
                "docker-compose",
                "run",
                "--rm",
                "-T",
                "django",
                "./manage.py",
                "drf_create_token",
                "admin@nil.nil",
            ],
            cwd=str(LOCAL_DOCKER_DIR),
            env=env,
            universal_newlines=True,
        )
        m = re.search(r"^Generated token (\w+) for user admin@nil.nil$", r, flags=re.M)
        if not m:
            raise RuntimeError(
                f"Could not extract Django auth token from drf_create_token output: {r!r}"
            )
        django_api_key = m[1]

        if create:
            run(
                ["docker-compose", "up", "-d", "django", "celery", "redirector"],
                cwd=str(LOCAL_DOCKER_DIR),
                env=env,
                check=True,
            )
            API_URL = known_instances["dandi-api-local-docker-tests"].api
            for _ in range(10):
                try:
                    requests.get(f"{API_URL}/dandisets/")
                except requests.ConnectionError:
                    sleep(1)
                else:
                    break
            else:
                raise RuntimeError("Django container did not start up in time")
        yield {"django_api_key": django_api_key}
    finally:
        if persist in (None, "0"):
            run(["docker-compose", "down", "-v"], cwd=str(LOCAL_DOCKER_DIR), check=True)


@dataclass
class DandiAPI:
    api_key: str
    client: DandiAPIClient
    instance: DandiInstance
    instance_id: str

    @property
    def api_url(self) -> str:
        url = self.instance.api
        assert isinstance(url, str)
        return url


@pytest.fixture(scope="session")
def local_dandi_api(docker_compose_setup: Dict[str, str]) -> Iterator[DandiAPI]:
    instance_id = "dandi-api-local-docker-tests"
    instance = known_instances[instance_id]
    api_key = docker_compose_setup["django_api_key"]
    with DandiAPIClient(api_url=instance.api, token=api_key) as client:
        yield DandiAPI(
            api_key=api_key,
            client=client,
            instance=instance,
            instance_id=instance_id,
        )


@dataclass
class SampleDandiset:
    api: DandiAPI
    dspath: Path
    dandiset: RemoteDandiset
    dandiset_id: str
    upload_kwargs: Dict[str, Any] = field(default_factory=dict)

    @property
    def client(self) -> DandiAPIClient:
        return self.api.client

    def upload(
        self, paths: Optional[List[Union[str, Path]]] = None, **kwargs: Any
    ) -> None:
        with pytest.MonkeyPatch().context() as m:
            m.setenv("DANDI_API_KEY", self.api.api_key)
            upload(
                paths=paths or [self.dspath],
                dandi_instance=self.api.instance_id,
                devel_debug=True,
                **{**self.upload_kwargs, **kwargs},
            )


@pytest.fixture()
def new_dandiset(
    local_dandi_api: DandiAPI,
    request: FixtureRequest,
    tmp_path_factory: pytest.TempPathFactory,
) -> SampleDandiset:
    d = local_dandi_api.client.create_dandiset(
        f"Sample Dandiset for {request.node.name}",
        # Minimal metadata needed to create a publishable Dandiset:
        {
            "description": "A test Dandiset",
            "license": ["spdx:CC0-1.0"],
            # The contributor needs to be given explicitly here or else it'll
            # be set based on the user account.  For the Docker Compose setup,
            # that would mean basing it on the admin user, whose name doesn't
            # validate under dandischema.
            "contributor": [
                {
                    "schemaKey": "Person",
                    "name": "Wodder, John",
                    "roleName": ["dcite:Author", "dcite:ContactPerson"],
                }
            ],
        },
    )
    dspath = tmp_path_factory.mktemp("dandiset")
    (dspath / dandiset_metadata_file).write_text(f"identifier: '{d.identifier}'\n")
    return SampleDandiset(
        api=local_dandi_api,
        dspath=dspath,
        dandiset=d,
        dandiset_id=d.identifier,
    )


@pytest.fixture()
def text_dandiset(new_dandiset: SampleDandiset) -> SampleDandiset:
    (new_dandiset.dspath / "file.txt").write_text("This is test text.\n")
    (new_dandiset.dspath / "subdir1").mkdir()
    (new_dandiset.dspath / "subdir1" / "apple.txt").write_text("Apple\n")
    (new_dandiset.dspath / "subdir2").mkdir()
    (new_dandiset.dspath / "subdir2" / "banana.txt").write_text("Banana\n")
    (new_dandiset.dspath / "subdir2" / "coconut.txt").write_text("Coconut\n")
    new_dandiset.upload_kwargs["allow_any_path"] = True
    new_dandiset.upload()
    return new_dandiset


@pytest.fixture()
def zarr_dandiset(new_dandiset: SampleDandiset) -> SampleDandiset:
    zarr.save(
        new_dandiset.dspath / "sample.zarr", np.arange(1000), np.arange(1000, 0, -1)
    )
    new_dandiset.upload()
    return new_dandiset


@pytest.fixture()
def video_files(tmp_path):
    video_paths = []
    import cv2

    video_path = tmp_path / "video_files"
    video_path.mkdir()
    for no in range(2):
        movie_file1 = video_path / f"test1_{no}.avi"
        movie_file2 = video_path / f"test2_{no}.avi"
        (nf, nx, ny) = (2, 2, 2)
        writer1 = cv2.VideoWriter(
            filename=str(movie_file1),
            apiPreference=None,
            fourcc=cv2.VideoWriter_fourcc(*"DIVX"),
            fps=25,
            frameSize=(ny, nx),
            params=None,
        )
        writer2 = cv2.VideoWriter(
            filename=str(movie_file2),
            apiPreference=None,
            fourcc=cv2.VideoWriter_fourcc(*"DIVX"),
            fps=25,
            frameSize=(ny, nx),
            params=None,
        )
        for k in range(nf):
            writer1.write(np.random.randint(0, 255, (nx, ny, 3)).astype("uint8"))
            writer2.write(np.random.randint(0, 255, (nx, ny, 3)).astype("uint8"))
        writer1.release()
        writer2.release()
        video_paths.append((movie_file1, movie_file2))
    return video_paths


def _create_nwb_files(video_list):
    base_path = video_list[0][0].parent.parent
    base_nwb_path = base_path / "nwbfiles"
    base_nwb_path.mkdir(parents=True, exist_ok=True)
    for no, vid_loc in enumerate(video_list):
        vid_1 = vid_loc[0]
        vid_2 = vid_loc[1]
        subject_id = f"mouse{no}"
        session_id = f"sessionid{no}"
        subject = Subject(
            subject_id=subject_id,
            species="Mus musculus",
            sex="M",
            description="lab mouse ",
        )
        device = Device(f"imaging_device_{no}")
        name = f"{vid_1.stem}_{no}"
        nwbfile = NWBFile(
            f"{name}{no}",
            "desc: contains movie for dandi .mp4 storage as external",
            datetime.now(tzlocal()),
            experimenter="Experimenter name",
            session_id=session_id,
            subject=subject,
            devices=[device],
        )

        image_series = ImageSeries(
            name=f"MouseWhiskers{no}",
            format="external",
            external_file=[str(vid_1), str(vid_2)],
            starting_frame=[0],
            starting_time=0.0,
            rate=150.0,
        )
        nwbfile.add_acquisition(image_series)

        nwbfile_path = base_nwb_path / f"{name}.nwb"
        with NWBHDF5IO(str(nwbfile_path), "w") as io:
            io.write(nwbfile)
    return base_nwb_path


@pytest.fixture()
def nwbfiles_video_unique(video_files):
    """Create nwbfiles linked with unique set of videos."""
    return _create_nwb_files(video_files)


@pytest.fixture()
def nwbfiles_video_common(video_files):
    """Create nwbfiles sharing video files."""
    video_list = [video_files[0], video_files[0]]
    return _create_nwb_files(video_list)


@pytest.fixture()
def tmp_home(
    monkeypatch: pytest.MonkeyPatch, tmp_path_factory: pytest.TempPathFactory
) -> Path:
    home = tmp_path_factory.mktemp("tmp_home")
    monkeypatch.setenv("HOME", str(home))
    monkeypatch.delenv("XDG_CACHE_HOME", raising=False)
    monkeypatch.delenv("XDG_CONFIG_DIRS", raising=False)
    monkeypatch.delenv("XDG_CONFIG_HOME", raising=False)
    monkeypatch.delenv("XDG_DATA_DIRS", raising=False)
    monkeypatch.delenv("XDG_DATA_HOME", raising=False)
    monkeypatch.delenv("XDG_RUNTIME_DIR", raising=False)
    monkeypatch.delenv("XDG_STATE_HOME", raising=False)
    monkeypatch.setenv("USERPROFILE", str(home))
    monkeypatch.setenv("LOCALAPPDATA", str(home))
    return home



from dataclasses import dataclass, field
from operator import attrgetter
from pathlib import Path
from typing import Iterable, Iterator, List, Optional, Tuple

import click

from .consts import DRAFT, dandiset_metadata_file
from .dandiapi import DandiAPIClient, RemoteAsset, RemoteDandiset
from .dandiarchive import BaseAssetIDURL, DandisetURL, ParsedDandiURL, parse_dandi_url
from .exceptions import NotFoundError
from .files import ZarrAsset
from .utils import get_instance, is_url


@dataclass
class Deleter:
    """
    Class for registering assets & Dandisets to delete and then deleting them
    """

    client: Optional[DandiAPIClient] = None
    dandiset: Optional[RemoteDandiset] = None
    #: Whether we are deleting an entire Dandiset (true) or just assets (false)
    deleting_dandiset: bool = False
    skip_missing: bool = False
    remote_assets: List[RemoteAsset] = field(default_factory=list)

    def __bool__(self) -> bool:
        return self.deleting_dandiset or bool(self.remote_assets)

    def set_dandiset(self, api_url: str, dandiset_id: str) -> bool:
        """
        Returns `False` if no action should be taken due to the Dandiset not
        existing
        """
        if self.client is None:
            # Strip the trailing slash so that dandi_authenticate can find the
            # URL in known_instances_rev:
            self.client = DandiAPIClient(api_url.rstrip("/"))
            self.client.dandi_authenticate()
            try:
                self.dandiset = self.client.get_dandiset(dandiset_id, DRAFT, lazy=False)
            except NotFoundError:
                if self.skip_missing:
                    return False
                else:
                    raise
        elif not is_same_url(self.client.api_url, api_url):
            raise ValueError("Cannot delete assets from multiple API instances at once")
        else:
            assert self.dandiset is not None
            if self.dandiset.identifier != dandiset_id:
                raise ValueError("Cannot delete assets from multiple Dandisets at once")
        return True

    def add_asset(self, asset: RemoteAsset) -> None:
        # Ensure the list is free of duplicates so that we don't try to delete
        # the same asset twice
        if not any(a.identifier == asset.identifier for a in self.remote_assets):
            self.remote_assets.append(asset)

    def register_dandiset(self, api_url: str, dandiset_id: str) -> None:
        if not self.set_dandiset(api_url, dandiset_id):
            return
        self.deleting_dandiset = True

    def register_asset(
        self, api_url: str, dandiset_id: str, version_id: str, asset_path: str
    ) -> None:
        if not self.set_dandiset(api_url, dandiset_id):
            return
        assert self.dandiset is not None
        try:
            asset = self.dandiset.get_asset_by_path(asset_path)
        except NotFoundError:
            if self.skip_missing:
                return
            else:
                raise NotFoundError(
                    f"Asset at path {asset_path!r} not found in Dandiset {dandiset_id}"
                )
        self.add_asset(asset)

    def register_asset_folder(
        self, api_url: str, dandiset_id: str, version_id: str, folder_path: str
    ) -> None:
        if not self.set_dandiset(api_url, dandiset_id):
            return
        any_assets = False
        assert self.dandiset is not None
        for asset in self.dandiset.get_assets_with_path_prefix(folder_path):
            self.add_asset(asset)
            any_assets = True
        if not any_assets and not self.skip_missing:
            raise NotFoundError(
                f"No assets under path {folder_path!r} found in Dandiset {dandiset_id}"
            )

    def register_assets_url(self, url: str, parsed_url: ParsedDandiURL) -> None:
        if isinstance(parsed_url, BaseAssetIDURL):
            raise ValueError("Cannot delete an asset identified by just an ID")
        assert parsed_url.dandiset_id is not None
        if not self.set_dandiset(parsed_url.api_url, parsed_url.dandiset_id):
            return
        any_assets = False
        assert self.client is not None
        for a in parsed_url.get_assets(self.client):
            assert isinstance(a, RemoteAsset)
            self.add_asset(a)
            any_assets = True
        if not any_assets and not self.skip_missing:
            raise NotFoundError(f"No assets found for {url}")

    def register_url(self, url: str) -> None:
        parsed_url = parse_dandi_url(url)
        if isinstance(parsed_url, DandisetURL):
            if parsed_url.version_id is not None:
                raise NotImplementedError(
                    "Dandi API server does not support deletion of individual"
                    " versions of a dandiset"
                )
            assert parsed_url.dandiset_id is not None
            self.register_dandiset(parsed_url.api_url, parsed_url.dandiset_id)
        else:
            if parsed_url.version_id is None:
                parsed_url.version_id = DRAFT
            self.register_assets_url(url, parsed_url)

    def register_local_path_equivalent(self, instance_name: str, filepath: str) -> None:
        instance = get_instance(instance_name)
        api_url = instance.api
        assert api_url is not None
        dandiset_id, asset_path = find_local_asset(filepath)
        if not self.set_dandiset(api_url, dandiset_id):
            return
        if asset_path.endswith("/"):
            self.register_asset_folder(api_url, dandiset_id, DRAFT, asset_path)
        else:
            self.register_asset(api_url, dandiset_id, DRAFT, asset_path)

    def confirm(self) -> bool:
        if self.dandiset is None:
            raise ValueError("confirm() called before registering anything to delete")
        if self.deleting_dandiset:
            msg = f"Delete Dandiset {self.dandiset.identifier}?"
        else:
            msg = (
                f"Delete {len(self.remote_assets)} assets on server from"
                f" Dandiset {self.dandiset.identifier}?"
            )
        return click.confirm(msg)

    def delete_dandiset(self) -> None:
        if self.deleting_dandiset:
            assert self.dandiset is not None
            self.dandiset.delete()
        else:
            raise RuntimeError(
                "delete_dandiset() called when Dandiset not registered for deletion"
            )

    def _process_asset(self, asset: RemoteAsset) -> Iterator[dict]:
        yield {"status": "Deleting"}
        try:
            asset.delete()
        except Exception as e:
            yield {"status": "Error", "message": f"{type(e).__name__}: {e}"}
        else:
            yield {"status": "Deleted"}

    def process_assets_pyout(self) -> Iterator[dict]:
        for asset in sorted(self.remote_assets, key=attrgetter("path")):
            yield {
                "path": asset.path,
                ("status", "message"): self._process_asset(asset),
            }

    def process_assets_debug(self) -> Iterator[Iterator[dict]]:
        for asset in sorted(self.remote_assets, key=attrgetter("path")):
            yield ({"path": asset.path, **d} for d in self._process_asset(asset))


def delete(
    paths: Iterable[str],
    dandi_instance: str = "dandi",
    devel_debug: bool = False,
    jobs: Optional[int] = None,
    force: bool = False,
    skip_missing: bool = False,
) -> None:
    deleter = Deleter(skip_missing=skip_missing)
    for p in paths:
        if is_url(p):
            deleter.register_url(p)
        else:
            deleter.register_local_path_equivalent(dandi_instance, p)
    if deleter and (force or deleter.confirm()):
        if deleter.deleting_dandiset:
            deleter.delete_dandiset()
        elif devel_debug:
            for gen in deleter.process_assets_debug():
                for r in gen:
                    print(r, flush=True)
        else:
            from .support import pyout as pyouts

            pyout_style = pyouts.get_style(hide_if_missing=False)
            rec_fields = ("path", "status", "message")
            out = pyouts.LogSafeTabular(
                style=pyout_style, columns=rec_fields, max_workers=jobs
            )
            with out:
                for r in deleter.process_assets_pyout():
                    out(r)


def find_local_asset(filepath: str) -> Tuple[str, str]:
    """
    Given a path to a local file, return the ID of the Dandiset in which it is
    located and the path to the file relative to the root of said Dandiset.  If
    the file is a directory, the path will end with a trailing slash.
    """
    from .dandiset import Dandiset

    path = Path(filepath).absolute()
    dandiset = Dandiset.find(path.parent)
    if dandiset is None:
        raise RuntimeError(
            f"Found no {dandiset_metadata_file} anywhere.  "
            "Use 'dandi download' or 'organize' first"
        )
    relpath = path.relative_to(dandiset.path).as_posix()
    if path.is_dir() and path.suffix not in ZarrAsset.EXTENSIONS:
        relpath += "/"
    return (dandiset.identifier, relpath)


def is_same_url(url1: str, url2: str) -> bool:
    # TODO: Use a real URL library like furl, purl, or yarl
    return url1.rstrip("/") == url2.rstrip("/")



"""
.. versionadded:: 0.36.0

This module defines functionality for working with local files & directories
(as opposed to remote resources on a DANDI Archive server) that are of interest
to DANDI.  The classes for such files & directories all inherit from
`DandiFile`, which has two immediate subclasses: `DandisetMetadataFile`, for
representing :file:`dandiset.yaml` files, and `LocalAsset`, for representing
files that can be uploaded as assets to DANDI Archive.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import deque
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from contextlib import closing
from dataclasses import dataclass, field, replace
from datetime import datetime
import os
from pathlib import Path
import re
from threading import Lock
from time import sleep
from typing import (
    Any,
    BinaryIO,
    ClassVar,
    Dict,
    Generator,
    Generic,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)
from xml.etree.ElementTree import fromstring

from dandischema.digests.dandietag import DandiETag
from dandischema.digests.zarr import get_checksum
from dandischema.models import BareAsset, CommonModel
from dandischema.models import Dandiset as DandisetMeta
from dandischema.models import DigestType, get_schema_version
from nwbinspector import Importance, inspect_nwb, load_config
from pydantic import ValidationError
import requests
import zarr

from . import get_logger
from .consts import (
    MAX_ZARR_DEPTH,
    VIDEO_FILE_EXTENSIONS,
    ZARR_MIME_TYPE,
    ZARR_UPLOAD_BATCH_SIZE,
    EmbargoStatus,
    dandiset_metadata_file,
)
from .dandiapi import (
    RemoteAsset,
    RemoteDandiset,
    RemoteZarrAsset,
    RemoteZarrEntry,
    RESTFullAPIClient,
)
from .exceptions import UnknownAssetError
from .metadata import get_default_metadata, nwb2asset
from .misctypes import DUMMY_DIGEST, BasePath, Digest, P
from .pynwb_utils import validate as pynwb_validate
from .support.digests import (
    get_dandietag,
    get_digest,
    get_zarr_checksum,
    md5file_nocache,
)
from .utils import chunked, pluralize, yaml_load

lgr = get_logger()

# TODO -- should come from schema.  This is just a simplistic example for now
_required_dandiset_metadata_fields = ["identifier", "name", "description"]


@dataclass  # type: ignore[misc]  # <https://github.com/python/mypy/issues/5374>
class DandiFile(ABC):
    """Abstract base class for local files & directories of interest to DANDI"""

    #: The path to the actual file or directory on disk
    filepath: Path

    @property
    def size(self) -> int:
        """The size of the file"""
        return os.path.getsize(self.filepath)

    @property
    def modified(self) -> datetime:
        """The time at which the file was last modified"""
        # TODO: Should this be overridden for LocalDirectoryAsset?
        return datetime.fromtimestamp(self.filepath.stat().st_mtime).astimezone()

    @abstractmethod
    def get_metadata(
        self,
        digest: Optional[Digest] = None,
        ignore_errors: bool = True,
    ) -> CommonModel:
        """Return the Dandi metadata for the file"""
        ...

    @abstractmethod
    def get_validation_errors(
        self,
        schema_version: Optional[str] = None,
        devel_debug: bool = False,
    ) -> List[str]:
        """
        Attempt to validate the file and return a list of errors encountered
        """
        ...


class DandisetMetadataFile(DandiFile):
    """Representation of a :file:`dandiset.yaml` file"""

    def get_metadata(
        self,
        digest: Optional[Digest] = None,
        ignore_errors: bool = True,
    ) -> DandisetMeta:
        """Return the Dandiset metadata inside the file"""
        with open(self.filepath) as f:
            meta = yaml_load(f, typ="safe")
        return DandisetMeta.unvalidated(**meta)

    # TODO: @validate_cache.memoize_path
    def get_validation_errors(
        self,
        schema_version: Optional[str] = None,
        devel_debug: bool = False,
    ) -> List[str]:
        with open(self.filepath) as f:
            meta = yaml_load(f, typ="safe")
        if schema_version is None:
            schema_version = meta.get("schemaVersion")
        if schema_version is None:
            return _check_required_fields(meta, _required_dandiset_metadata_fields)
        else:
            current_version = get_schema_version()
            if schema_version != current_version:
                raise ValueError(
                    f"Unsupported schema version: {schema_version}; expected {current_version}"
                )
            try:
                DandisetMeta(**meta)
            except ValidationError as e:
                if devel_debug:
                    raise
                lgr.warning(
                    "Validation error for %s: %s",
                    self.filepath,
                    e,
                    extra={"validating": True},
                )
                return [str(e)]
            except Exception as e:
                if devel_debug:
                    raise
                lgr.warning(
                    "Unexpected validation error for %s: %s",
                    self.filepath,
                    e,
                    extra={"validating": True},
                )
                return [f"Failed to initialize Dandiset meta: {e}"]
            return []


@dataclass  # type: ignore[misc]  # <https://github.com/python/mypy/issues/5374>
class LocalAsset(DandiFile):
    """
    Representation of a file or directory that can be uploaded to a DANDI
    Archive as an asset of a Dandiset
    """

    #: The forward-slash-separated path to the asset within its local Dandiset
    #: (i.e., relative to the Dandiset's root)
    path: str

    @abstractmethod
    def get_digest(self) -> Digest:
        """
        Calculate a DANDI etag digest for the asset using the appropriate
        algorithm for its type
        """
        ...

    @abstractmethod
    def get_metadata(
        self,
        digest: Optional[Digest] = None,
        ignore_errors: bool = True,
    ) -> BareAsset:
        """Return the Dandi metadata for the asset"""
        ...

    # TODO: @validate_cache.memoize_path
    def get_validation_errors(
        self,
        schema_version: Optional[str] = None,
        devel_debug: bool = False,
    ) -> List[str]:
        if schema_version is not None:
            current_version = get_schema_version()
            if schema_version != current_version:
                raise ValueError(
                    f"Unsupported schema version: {schema_version}; expected {current_version}"
                )
            try:
                asset = self.get_metadata(digest=DUMMY_DIGEST)
                BareAsset(**asset.dict())
            except ValidationError as e:
                if devel_debug:
                    raise
                lgr.warning(
                    "Validation error for %s: %s",
                    self.filepath,
                    e,
                    extra={"validating": True},
                )
                return [str(e)]
            except Exception as e:
                if devel_debug:
                    raise
                lgr.warning(
                    "Unexpected validation error for %s: %s",
                    self.filepath,
                    e,
                    extra={"validating": True},
                )
                return [f"Failed to read metadata: {e}"]
            return []
        else:
            # TODO: Do something else?
            return []

    def upload(
        self,
        dandiset: RemoteDandiset,
        metadata: Dict[str, Any],
        jobs: Optional[int] = None,
        replacing: Optional[RemoteAsset] = None,
    ) -> RemoteAsset:
        """
        Upload the file as an asset with the given metadata to the given
        Dandiset and return the resulting asset.  Blocks until the upload is
        complete.

        :param RemoteDandiset dandiset:
            the Dandiset to which the file will be uploaded
        :param dict metadata:
            Metadata for the uploaded asset.  The "path" field will be set to
            the value of the instance's ``path`` attribute if no such field is
            already present.
        :param int jobs: Number of threads to use for uploading; defaults to 5
        :param RemoteAsset replacing:
            If set, replace the given asset, which must have the same path as
            the new asset
        :rtype: RemoteAsset
        """
        for status in self.iter_upload(
            dandiset, metadata, jobs=jobs, replacing=replacing
        ):
            if status["status"] == "done":
                a = status["asset"]
                assert isinstance(a, RemoteAsset)
                return a
        raise AssertionError("iter_upload() finished without returning 'done'")

    @abstractmethod
    def iter_upload(
        self,
        dandiset: RemoteDandiset,
        metadata: Dict[str, Any],
        jobs: Optional[int] = None,
        replacing: Optional[RemoteAsset] = None,
    ) -> Iterator[dict]:
        """
        Upload the asset with the given metadata to the given Dandiset,
        returning a generator of status `dict`\\s.

        :param RemoteDandiset dandiset:
            the Dandiset to which the asset will be uploaded
        :param dict metadata:
            Metadata for the uploaded asset.  The "path" field will be set to
            the value of the instance's ``path`` attribute if no such field is
            already present.
        :param int jobs: Number of threads to use for uploading; defaults to 5
        :param RemoteAsset replacing:
            If set, replace the given asset, which must have the same path as
            the new asset
        :returns:
            A generator of `dict`\\s containing at least a ``"status"`` key.
            Upon successful upload, the last `dict` will have a status of
            ``"done"`` and an ``"asset"`` key containing the resulting
            `RemoteAsset`.
        """
        ...


class LocalFileAsset(LocalAsset):
    """
    Representation of a regular file that can be uploaded to a DANDI Archive as
    an asset of a Dandiset
    """

    EXTENSIONS: ClassVar[List[str]] = []

    def get_metadata(
        self,
        digest: Optional[Digest] = None,
        ignore_errors: bool = True,
    ) -> BareAsset:
        metadata = get_default_metadata(self.filepath, digest=digest)
        metadata.path = self.path
        return metadata

    def get_digest(self) -> Digest:
        """Calculate a dandi-etag digest for the asset"""
        value = get_digest(self.filepath, digest="dandi-etag")
        return Digest.dandi_etag(value)

    def iter_upload(
        self,
        dandiset: RemoteDandiset,
        metadata: Dict[str, Any],
        jobs: Optional[int] = None,
        replacing: Optional[RemoteAsset] = None,
    ) -> Iterator[dict]:
        """
        Upload the file as an asset with the given metadata to the given
        Dandiset, returning a generator of status `dict`\\s.

        :param RemoteDandiset dandiset:
            the Dandiset to which the file will be uploaded
        :param dict metadata:
            Metadata for the uploaded asset.  The "path" field will be set to
            the value of the instance's ``path`` attribute if no such field is
            already present.
        :param int jobs: Number of threads to use for uploading; defaults to 5
        :param RemoteAsset replacing:
            If set, replace the given asset, which must have the same path as
            the new asset
        :returns:
            A generator of `dict`\\s containing at least a ``"status"`` key.
            Upon successful upload, the last `dict` will have a status of
            ``"done"`` and an ``"asset"`` key containing the resulting
            `RemoteAsset`.
        """
        asset_path = metadata.setdefault("path", self.path)
        client = dandiset.client
        yield {"status": "calculating etag"}
        etagger = get_dandietag(self.filepath)
        filetag = etagger.as_str()
        lgr.debug("Calculated dandi-etag of %s for %s", filetag, self.filepath)
        digest = metadata.get("digest", {})
        if "dandi:dandi-etag" in digest:
            if digest["dandi:dandi-etag"] != filetag:
                raise RuntimeError(
                    f"{self.filepath}: File etag changed; was originally"
                    f" {digest['dandi:dandi-etag']} but is now {filetag}"
                )
        yield {"status": "initiating upload"}
        lgr.debug("%s: Beginning upload", asset_path)
        total_size = self.size
        try:
            resp = client.post(
                "/uploads/initialize/",
                json={
                    "contentSize": total_size,
                    "digest": {
                        "algorithm": "dandi:dandi-etag",
                        "value": filetag,
                    },
                    "dandiset": dandiset.identifier,
                },
            )
        except requests.HTTPError as e:
            if e.response.status_code == 409:
                lgr.debug("%s: Blob already exists on server", asset_path)
                blob_id = e.response.headers["Location"]
            else:
                raise
        else:
            upload_id = resp["upload_id"]
            parts = resp["parts"]
            if len(parts) != etagger.part_qty:
                raise RuntimeError(
                    f"Server and client disagree on number of parts for upload;"
                    f" server says {len(parts)}, client says {etagger.part_qty}"
                )
            parts_out = []
            bytes_uploaded = 0
            lgr.debug("Uploading %s in %d parts", self.filepath, len(parts))
            with RESTFullAPIClient("http://nil.nil") as storage:
                with self.filepath.open("rb") as fp:
                    with ThreadPoolExecutor(max_workers=jobs or 5) as executor:
                        lock = Lock()
                        futures = [
                            executor.submit(
                                _upload_blob_part,
                                storage_session=storage,
                                fp=fp,
                                lock=lock,
                                etagger=etagger,
                                asset_path=asset_path,
                                part=part,
                            )
                            for part in parts
                        ]
                        for fut in as_completed(futures):
                            out_part = fut.result()
                            bytes_uploaded += out_part["size"]
                            yield {
                                "status": "uploading",
                                "upload": 100 * bytes_uploaded / total_size,
                                "current": bytes_uploaded,
                            }
                            parts_out.append(out_part)
                lgr.debug("%s: Completing upload", asset_path)
                resp = client.post(
                    f"/uploads/{upload_id}/complete/",
                    json={"parts": parts_out},
                )
                lgr.debug(
                    "%s: Announcing completion to %s",
                    asset_path,
                    resp["complete_url"],
                )
                r = storage.post(
                    resp["complete_url"], data=resp["body"], json_resp=False
                )
                lgr.debug(
                    "%s: Upload completed. Response content: %s",
                    asset_path,
                    r.content,
                )
                rxml = fromstring(r.text)
                m = re.match(r"\{.+?\}", rxml.tag)
                ns = m.group(0) if m else ""
                final_etag = rxml.findtext(f"{ns}ETag")
                if final_etag is not None:
                    final_etag = final_etag.strip('"')
                    if final_etag != filetag:
                        raise RuntimeError(
                            "Server and client disagree on final ETag of uploaded file;"
                            f" server says {final_etag}, client says {filetag}"
                        )
                # else: Error? Warning?
                resp = client.post(f"/uploads/{upload_id}/validate/")
                blob_id = resp["blob_id"]
        lgr.debug("%s: Assigning asset blob to dandiset & version", asset_path)
        yield {"status": "producing asset"}
        if replacing is not None:
            lgr.debug("%s: Replacing pre-existing asset", asset_path)
            r = client.put(
                replacing.api_path,
                json={"metadata": metadata, "blob_id": blob_id},
            )
        else:
            r = client.post(
                f"{dandiset.version_api_path}assets/",
                json={"metadata": metadata, "blob_id": blob_id},
            )
        a = RemoteAsset.from_data(dandiset, r)
        lgr.info("%s: Asset successfully uploaded", asset_path)
        yield {"status": "done", "asset": a}


class NWBAsset(LocalFileAsset):
    """Representation of a local NWB file"""

    EXTENSIONS: ClassVar[List[str]] = [".nwb"]

    def get_metadata(
        self,
        digest: Optional[Digest] = None,
        ignore_errors: bool = True,
    ) -> BareAsset:
        try:
            metadata = nwb2asset(self.filepath, digest=digest)
        except Exception as e:
            lgr.warning(
                "Failed to extract NWB metadata from %s: %s: %s",
                self.filepath,
                type(e).__name__,
                str(e),
            )
            if ignore_errors:
                metadata = get_default_metadata(self.filepath, digest=digest)
            else:
                raise
        metadata.path = self.path
        return metadata

    # TODO: @validate_cache.memoize_path
    def get_validation_errors(
        self,
        schema_version: Optional[str] = None,
        devel_debug: bool = False,
    ) -> List[str]:
        errors: List[str] = pynwb_validate(self.filepath, devel_debug=devel_debug)
        if schema_version is not None:
            errors.extend(
                super().get_validation_errors(
                    schema_version=schema_version, devel_debug=devel_debug
                )
            )
        else:
            # make sure that we have some basic metadata fields we require
            try:
                errors.extend(
                    [
                        error.message
                        for error in inspect_nwb(
                            nwbfile_path=self.filepath,
                            skip_validate=True,
                            config=load_config(filepath_or_keyword="dandi"),
                            importance_threshold=Importance.CRITICAL,
                        )
                    ]
                )
            except Exception as e:
                if devel_debug:
                    raise
                lgr.warning(
                    "Failed to inspect NWBFile in %s: %s",
                    self.filepath,
                    e,
                    extra={"validating": True},
                )
                errors.append(f"Failed to inspect NWBFile: {e}")
        return errors


class VideoAsset(LocalFileAsset):
    EXTENSIONS: ClassVar[List[str]] = VIDEO_FILE_EXTENSIONS


class GenericAsset(LocalFileAsset):
    """
    Representation of a generic regular file, one that is not of any known type
    """

    EXTENSIONS: ClassVar[List[str]] = []


class LocalDirectoryAsset(LocalAsset, Generic[P]):
    """
    Representation of a directory that can be uploaded to a DANDI Archive as
    a single asset of a Dandiset.  It is generic in ``P``, bound to
    `dandi.misctypes.BasePath`.
    """

    EXTENSIONS: ClassVar[List[str]] = []

    @property
    @abstractmethod
    def filetree(self) -> P:
        """
        The path object for the root of the hierarchy of files within the
        directory
        """
        ...

    def iterfiles(self, include_dirs: bool = False) -> Iterator[P]:
        """Yield all files within the directory"""
        dirs = deque([self.filetree])
        while dirs:
            for p in dirs.popleft().iterdir():
                if p.is_dir():
                    dirs.append(p)
                    if include_dirs:
                        yield p
                else:
                    yield p

    @property
    def size(self) -> int:
        """The total size of the files in the directory"""
        return sum(p.size for p in self.iterfiles())


@dataclass
class LocalZarrEntry(BasePath):
    """A file or directory within a `ZarrAsset`"""

    #: The path to the actual file or directory on disk
    filepath: Path
    #: The path to the root of the Zarr file tree
    zarr_basepath: Path

    def _get_subpath(self, name: str) -> LocalZarrEntry:
        if not name or "/" in name:
            raise ValueError(f"Invalid path component: {name!r}")
        elif name == ".":
            return self
        elif name == "..":
            return self.parent
        else:
            return replace(
                self, filepath=self.filepath / name, parts=self.parts + (name,)
            )

    @property
    def parent(self) -> LocalZarrEntry:
        if self.is_root():
            return self
        else:
            return replace(self, filepath=self.filepath.parent, parts=self.parts[:-1])

    def exists(self) -> bool:
        return self.filepath.exists()

    def is_file(self) -> bool:
        return self.filepath.is_file()

    def is_dir(self) -> bool:
        return self.filepath.is_dir()

    def iterdir(self) -> Iterator[LocalZarrEntry]:
        for p in self.filepath.iterdir():
            if p.is_dir() and not any(p.iterdir()):
                # Ignore empty directories
                continue
            yield self._get_subpath(p.name)

    def get_digest(self) -> Digest:
        """
        Calculate the DANDI etag digest for the entry.  If the entry is a
        directory, the algorithm will be the Dandi Zarr checksum algorithm; if
        it is a file, it will be MD5.
        """
        if self.is_dir():
            return Digest.dandi_zarr(get_zarr_checksum(self.filepath))
        else:
            return Digest(
                algorithm=DigestType.md5, value=get_digest(self.filepath, "md5")
            )

    @property
    def size(self) -> int:
        """
        The size of the entry.  For a directory, this is the total size of all
        entries within it.
        """
        if self.is_dir():
            return sum(p.size for p in self.iterdir())
        else:
            return os.path.getsize(self.filepath)

    @property
    def modified(self) -> datetime:
        """The time at which the entry was last modified"""
        # TODO: Should this be overridden for directories?
        return datetime.fromtimestamp(self.filepath.stat().st_mtime).astimezone()


@dataclass
class ZarrStat:
    """Details about a Zarr asset"""

    #: The total size of the asset
    size: int
    #: The Dandi Zarr checksum of the asset
    digest: Digest
    #: A list of all files in the asset in unspecified order
    files: List[LocalZarrEntry]


class ZarrAsset(LocalDirectoryAsset[LocalZarrEntry]):
    """Representation of a local Zarr directory"""

    EXTENSIONS: ClassVar[List[str]] = [".ngff", ".zarr"]

    @property
    def filetree(self) -> LocalZarrEntry:
        """
        The `LocalZarrEntry` for the root of the hierarchy of files within the
        Zarr asset
        """
        return LocalZarrEntry(
            filepath=self.filepath, zarr_basepath=self.filepath, parts=()
        )

    def stat(self) -> ZarrStat:
        """Return various details about the Zarr asset"""

        def dirstat(dirpath: LocalZarrEntry) -> ZarrStat:
            size = 0
            dir_md5s = {}
            file_md5s = {}
            files = []
            for p in dirpath.iterdir():
                if p.is_dir():
                    st = dirstat(p)
                    size += st.size
                    dir_md5s[str(p)] = (st.digest.value, st.size)
                    files.extend(st.files)
                else:
                    size += p.size
                    file_md5s[str(p)] = (md5file_nocache(p.filepath), p.size)
                    files.append(p)
            return ZarrStat(
                size=size,
                digest=Digest.dandi_zarr(get_checksum(file_md5s, dir_md5s)),
                files=files,
            )

        return dirstat(self.filetree)

    def get_digest(self) -> Digest:
        """Calculate a dandi-zarr-checksum digest for the asset"""
        return Digest.dandi_zarr(get_zarr_checksum(self.filepath))

    def get_metadata(
        self,
        digest: Optional[Digest] = None,
        ignore_errors: bool = True,
    ) -> BareAsset:
        metadata = get_default_metadata(self.filepath, digest=digest)
        metadata.encodingFormat = ZARR_MIME_TYPE
        metadata.path = self.path
        return metadata

    def get_validation_errors(
        self,
        schema_version: Optional[str] = None,
        devel_debug: bool = False,
    ) -> List[str]:
        try:
            data = zarr.open(self.filepath)
        except Exception as e:
            if devel_debug:
                raise
            lgr.warning(
                "Error opening %s: %s: %s",
                self.filepath,
                type(e).__name__,
                e,
                extra={"validating": True},
            )
            return [str(e)]
        if isinstance(data, zarr.Group) and not data:
            msg = "Zarr group is empty"
            if devel_debug:
                raise ValueError(msg)
            lgr.warning("%s: %s", self.filepath, msg, extra={"validating": True})
            return [msg]
        try:
            next(self.filepath.glob(f"*{os.sep}" + os.sep.join(["*"] * MAX_ZARR_DEPTH)))
        except StopIteration:
            pass
        else:
            msg = f"Zarr directory tree more than {MAX_ZARR_DEPTH} directories deep"
            if devel_debug:
                raise ValueError(msg)
            lgr.warning("%s: %s", self.filepath, msg, extra={"validating": True})
            return [msg]
        # TODO: Should this be appended to the above errors?
        return super().get_validation_errors(
            schema_version=schema_version, devel_debug=devel_debug
        )

    def iter_upload(
        self,
        dandiset: RemoteDandiset,
        metadata: Dict[str, Any],
        jobs: Optional[int] = None,
        replacing: Optional[RemoteAsset] = None,
    ) -> Iterator[dict]:
        """
        Upload the Zarr directory as an asset with the given metadata to the
        given Dandiset, returning a generator of status `dict`\\s.

        :param RemoteDandiset dandiset:
            the Dandiset to which the Zarr will be uploaded
        :param dict metadata:
            Metadata for the uploaded asset.  The "path" field will be set to
            the value of the instance's ``path`` attribute if no such field is
            already present.
        :param int jobs: Number of threads to use for uploading; defaults to 5
        :param RemoteAsset replacing:
            If set, replace the given asset, which must have the same path as
            the new asset; if the old asset is a Zarr, the Zarr will be updated
            & reused for the new asset
        :returns:
            A generator of `dict`\\s containing at least a ``"status"`` key.
            Upon successful upload, the last `dict` will have a status of
            ``"done"`` and an ``"asset"`` key containing the resulting
            `RemoteAsset`.
        """
        # So that older clients don't get away with doing the wrong thing once
        # Zarr upload to embargoed Dandisets is implemented in the API:
        if dandiset.embargo_status is EmbargoStatus.EMBARGOED:
            raise NotImplementedError(
                "Uploading Zarr assets to embargoed Dandisets is currently not implemented"
            )
        asset_path = metadata.setdefault("path", self.path)
        client = dandiset.client
        lgr.debug("%s: Producing asset", asset_path)
        yield {"status": "producing asset"}
        old_zarr_entries: Dict[str, RemoteZarrEntry] = {}
        if replacing is not None:
            lgr.debug("%s: Replacing pre-existing asset", asset_path)
            if isinstance(replacing, RemoteZarrAsset):
                lgr.debug(
                    "%s: Pre-existing asset is a Zarr; reusing & updating", asset_path
                )
                zarr_id = replacing.zarr
                old_zarr_entries = {
                    str(e): e for e in replacing.iterfiles(include_dirs=True)
                }
            else:
                lgr.debug(
                    "%s: Pre-existing asset is not a Zarr; minting new Zarr", asset_path
                )
                r = client.post(
                    "/zarr/", json={"name": asset_path, "dandiset": dandiset.identifier}
                )
                zarr_id = r["zarr_id"]
            r = client.put(
                replacing.api_path,
                json={"metadata": metadata, "zarr_id": zarr_id},
            )
        else:
            lgr.debug("%s: Minting new Zarr", asset_path)
            r = client.post(
                "/zarr/", json={"name": asset_path, "dandiset": dandiset.identifier}
            )
            zarr_id = r["zarr_id"]
            r = client.post(
                f"{dandiset.version_api_path}assets/",
                json={"metadata": metadata, "zarr_id": zarr_id},
            )
        a = RemoteAsset.from_data(dandiset, r)
        assert isinstance(a, RemoteZarrAsset)

        total_size = 0
        to_upload = EntryUploadTracker()
        if old_zarr_entries:
            to_delete: List[RemoteZarrEntry] = []
            digesting: List[Future[Optional[Tuple[LocalZarrEntry, str]]]] = []
            yield {"status": "comparing against remote Zarr"}
            with ThreadPoolExecutor(max_workers=jobs or 5) as executor:
                for local_entry in self.iterfiles():
                    total_size += local_entry.size
                    try:
                        remote_entry = old_zarr_entries.pop(str(local_entry))
                    except KeyError:
                        for pp in local_entry.parents:
                            pps = str(pp)
                            if pps in old_zarr_entries:
                                if old_zarr_entries[pps].is_file():
                                    lgr.debug(
                                        "%s: Parent path %s of file %s"
                                        " corresponds to a remote file;"
                                        " deleting remote",
                                        asset_path,
                                        pps,
                                        local_entry,
                                    )
                                    to_delete.append(old_zarr_entries.pop(pps))
                                break
                        lgr.debug(
                            "%s: Path %s not present in remote Zarr; uploading",
                            asset_path,
                            local_entry,
                        )
                        to_upload.register(local_entry)
                    else:
                        if remote_entry.is_dir():
                            lgr.debug(
                                "%s: Path %s of local file is a directory in"
                                " remote Zarr; deleting remote & re-uploading",
                                asset_path,
                                local_entry,
                            )
                            eprefix = str(remote_entry) + "/"
                            sub_e = [
                                (k, v)
                                for k, v in old_zarr_entries.items()
                                if k.startswith(eprefix)
                            ]
                            for k, v in sub_e:
                                old_zarr_entries.pop(k)
                                to_delete.append(v)
                            to_upload.register(local_entry)
                        else:
                            digesting.append(
                                executor.submit(
                                    _cmp_digests,
                                    asset_path,
                                    local_entry,
                                    remote_entry.get_digest().value,
                                )
                            )
                for dgstfut in as_completed(digesting):
                    try:
                        item = dgstfut.result()
                    except Exception:
                        for d in digesting:
                            d.cancel()
                        raise
                    else:
                        if item is not None:
                            local_entry, local_digest = item
                            to_upload.register(local_entry, local_digest)
            if to_delete:
                a.rmfiles(to_delete, reingest=False)
        else:
            yield {"status": "traversing local Zarr"}
            for local_entry in self.iterfiles():
                total_size += local_entry.size
                to_upload.register(local_entry)
        yield {"status": "initiating upload", "size": total_size}
        lgr.debug("%s: Beginning upload", asset_path)
        bytes_uploaded = 0
        with RESTFullAPIClient(
            "http://nil.nil",
            headers={"X-Amz-ACL": "bucket-owner-full-control"},
        ) as storage, closing(to_upload.get_items()) as upload_items:
            for i, upload_body in enumerate(
                chunked(upload_items, ZARR_UPLOAD_BATCH_SIZE), start=1
            ):
                lgr.debug(
                    "%s: Uploading Zarr file batch #%d (%s)",
                    asset_path,
                    i,
                    pluralize(len(upload_body), "file"),
                )
                r = client.post(f"/zarr/{zarr_id}/upload/", json=upload_body)
                with ThreadPoolExecutor(max_workers=jobs or 5) as executor:
                    futures = [
                        executor.submit(
                            _upload_zarr_file,
                            storage_session=storage,
                            path=self.filepath / upspec["path"],
                            upload_url=upspec["upload_url"],
                        )
                        for upspec in r
                    ]
                    for fut in as_completed(futures):
                        try:
                            size = fut.result()
                        except Exception as e:
                            lgr.debug(
                                "Error uploading zarr: %s: %s", type(e).__name__, e
                            )
                            lgr.debug("Cancelling upload")
                            for f in futures:
                                f.cancel()
                            executor.shutdown()
                            client.delete(f"/zarr/{zarr_id}/upload/")
                            raise
                        else:
                            bytes_uploaded += size
                            yield {
                                "status": "uploading",
                                "upload": 100 * bytes_uploaded / to_upload.total_size,
                                "current": bytes_uploaded,
                            }
                lgr.debug("%s: Completing upload of batch #%d", asset_path, i)
                client.post(f"/zarr/{zarr_id}/upload/complete/")
        lgr.debug("%s: All files uploaded", asset_path)
        old_zarr_files = [e for e in old_zarr_entries.values() if e.is_file()]
        if old_zarr_files:
            yield {"status": "deleting extra remote files"}
            lgr.debug(
                "%s: Deleting %s in remote Zarr not present locally",
                asset_path,
                pluralize(len(old_zarr_files), "file"),
            )
            a.rmfiles(old_zarr_files, reingest=False)
        lgr.debug("%s: Waiting for server to calculate Zarr checksum", asset_path)
        yield {"status": "server calculating checksum"}
        client.post(f"/zarr/{zarr_id}/ingest/")
        while True:
            sleep(2)
            r = client.get(f"/zarr/{zarr_id}/")
            if r["status"] == "Complete":
                break
        lgr.info("%s: Asset successfully uploaded", asset_path)
        yield {"status": "done", "asset": a}


def find_dandi_files(
    *paths: Union[str, Path],
    dandiset_path: Optional[Union[str, Path]] = None,
    allow_all: bool = False,
    include_metadata: bool = False,
) -> Iterator[DandiFile]:
    """
    Yield all DANDI files at or under the paths in ``paths`` (which may be
    either files or directories).  Files & directories whose names start with a
    period are ignored.  Directories are only included in the return value if
    they are of a type represented by a `LocalDirectoryAsset` subclass, in
    which case they are not recursed into.

    :param dandiset_path:
        The path to the root of the Dandiset in which the paths are located.
        All paths in ``paths`` must be equal to or subpaths of
        ``dandiset_path``.  If `None`, then the Dandiset path for each asset
        found is implicitly set to the parent directory.
    :param allow_all:
        If true, unrecognized assets and the Dandiset's :file:`dandiset.yaml`
        file are returned as `GenericAsset` and `DandisetMetadataFile`
        instances, respectively.  If false, they are not returned at all.
    :param include_metadata:
        If true, the Dandiset's :file:`dandiset.yaml` file is returned as a
        `DandisetMetadataFile` instance.  If false, it is not returned at all
        (unless ``allow_all`` is true).
    """

    path_queue: deque[Path] = deque()
    for p in map(Path, paths):
        if dandiset_path is not None:
            try:
                p.relative_to(dandiset_path)
            except ValueError:
                raise ValueError(
                    "Path {str(p)!r} is not inside Dandiset path {str(dandiset_path)!r}"
                )
        path_queue.append(p)
    while path_queue:
        p = path_queue.popleft()
        if p.name.startswith("."):
            continue
        if p.is_dir():
            if p.is_symlink():
                lgr.warning("%s: Ignoring unsupported symbolic link to directory", p)
            elif dandiset_path is not None and p == Path(dandiset_path):
                path_queue.extend(p.iterdir())
            elif any(p.iterdir()):
                try:
                    df = dandi_file(p, dandiset_path)
                except UnknownAssetError:
                    path_queue.extend(p.iterdir())
                else:
                    yield df
        else:
            df = dandi_file(p, dandiset_path)
            if isinstance(df, GenericAsset) and not allow_all:
                pass
            elif isinstance(df, DandisetMetadataFile) and not (
                allow_all or include_metadata
            ):
                pass
            else:
                yield df


def dandi_file(
    filepath: Union[str, Path], dandiset_path: Optional[Union[str, Path]] = None
) -> DandiFile:
    """
    Return a `DandiFile` instance of the appropriate type for the file at
    ``filepath`` inside the Dandiset rooted at ``dandiset_path``.  If
    ``dandiset_path`` is not set, it will default to ``filepath``'s parent
    directory.

    If ``filepath`` is a directory, it must be of a type represented by a
    `LocalDirectoryAsset` subclass; otherwise, an `UnknownAssetError` exception
    will be raised.

    A regular file named :file:`dandiset.yaml` will only be represented by a
    `DandisetMetadataFile` instance if it is at the root of the Dandiset.

    A regular file that is not of a known type will be represented by a
    `GenericAsset` instance.
    """
    filepath = Path(filepath)
    if dandiset_path is not None:
        path = filepath.relative_to(dandiset_path).as_posix()
        if path == ".":
            raise ValueError("Dandi file path cannot equal Dandiset path")
    else:
        path = filepath.name
    if filepath.is_dir():
        if not any(filepath.iterdir()):
            raise UnknownAssetError("Empty directories cannot be assets")
        for dirclass in LocalDirectoryAsset.__subclasses__():
            if filepath.suffix in dirclass.EXTENSIONS:
                return dirclass(filepath=filepath, path=path)  # type: ignore[abstract]
        raise UnknownAssetError(
            f"Directory has unrecognized suffix {filepath.suffix!r}"
        )
    elif path == dandiset_metadata_file:
        return DandisetMetadataFile(filepath=filepath)
    else:
        for fileclass in LocalFileAsset.__subclasses__():
            if filepath.suffix in fileclass.EXTENSIONS:
                return fileclass(filepath=filepath, path=path)  # type: ignore[abstract]
        return GenericAsset(filepath=filepath, path=path)


def _upload_blob_part(
    storage_session: RESTFullAPIClient,
    fp: BinaryIO,
    lock: Lock,
    etagger: DandiETag,
    asset_path: str,
    part: dict,
) -> dict:
    etag_part = etagger.get_part(part["part_number"])
    if part["size"] != etag_part.size:
        raise RuntimeError(
            f"Server and client disagree on size of upload part"
            f" {part['part_number']}; server says {part['size']},"
            f" client says {etag_part.size}"
        )
    with lock:
        fp.seek(etag_part.offset)
        chunk = fp.read(part["size"])
    if len(chunk) != part["size"]:
        raise RuntimeError(
            f"End of file {fp.name} reached unexpectedly early:"
            f" read {len(chunk)} bytes of out of an expected {part['size']}"
        )
    lgr.debug(
        "%s: Uploading part %d/%d (%d bytes)",
        asset_path,
        part["part_number"],
        etagger.part_qty,
        part["size"],
    )
    r = storage_session.put(
        part["upload_url"],
        data=chunk,
        json_resp=False,
        retry_statuses=[500],
    )
    server_etag = r.headers["ETag"].strip('"')
    lgr.debug(
        "%s: Part upload finished ETag=%s Content-Length=%s",
        asset_path,
        server_etag,
        r.headers.get("Content-Length"),
    )
    client_etag = etagger.get_part_etag(etag_part)
    if server_etag != client_etag:
        raise RuntimeError(
            f"Server and client disagree on ETag of upload part"
            f" {part['part_number']}; server says"
            f" {server_etag}, client says {client_etag}"
        )
    return {
        "part_number": part["part_number"],
        "size": part["size"],
        "etag": server_etag,
    }


def _upload_zarr_file(
    storage_session: RESTFullAPIClient, path: Path, upload_url: str
) -> int:
    with path.open("rb") as fp:
        storage_session.put(upload_url, data=fp, json_resp=False)
    return path.stat().st_size


def _check_required_fields(d: dict, required: List[str]) -> List[str]:
    errors: List[str] = []
    for f in required:
        v = d.get(f, None)
        if not v or (isinstance(v, str) and not v.strip()):
            errors += [f"Required field {f!r} has no value"]
        if v in ("REQUIRED", "PLACEHOLDER"):
            errors += [f"Required field {f!r} has value {v!r}"]
    return errors


@dataclass
class EntryUploadTracker:
    """
    Class for keeping track of `LocalZarrEntry` instances to upload

    :meta private:
    """

    total_size: int = 0
    digested_entries: List[Tuple[LocalZarrEntry, str]] = field(default_factory=list)
    fresh_entries: List[LocalZarrEntry] = field(default_factory=list)

    def register(self, e: LocalZarrEntry, digest: Optional[str] = None) -> None:
        if digest is not None:
            self.digested_entries.append((e, digest))
        else:
            self.fresh_entries.append(e)
        self.total_size += e.size

    @staticmethod
    def _mkitem(e: LocalZarrEntry) -> dict:
        digest = md5file_nocache(e.filepath)
        return {"path": str(e), "etag": digest}

    def get_items(self, jobs: int = 5) -> Generator[dict, None, None]:
        # Note: In order for the ThreadPoolExecutor to be closed if an error
        # occurs during upload, the method must be used like this:
        #
        #     with contextlib.closing(to_upload.get_items()) as upload_items:
        #         for item in upload_items:
        #             ...
        for e, digest in self.digested_entries:
            yield {"path": str(e), "etag": digest}
        if not self.fresh_entries:
            return
        with ThreadPoolExecutor(max_workers=jobs) as executor:
            futures = [executor.submit(self._mkitem, e) for e in self.fresh_entries]
            for fut in as_completed(futures):
                try:
                    yield fut.result()
                # Use BaseException to also catch GeneratorExit thrown by
                # closing()
                except BaseException:
                    for f in futures:
                        f.cancel()
                    raise


def _cmp_digests(
    asset_path: str, local_entry: LocalZarrEntry, remote_digest: str
) -> Optional[Tuple[LocalZarrEntry, str]]:
    local_digest = md5file_nocache(local_entry.filepath)
    if local_digest != remote_digest:
        lgr.debug(
            "%s: Path %s in Zarr differs from local file; re-uploading",
            asset_path,
            local_entry,
        )
        return (local_entry, local_digest)
    else:
        lgr.debug("%s: File %s already on server; skipping", asset_path, local_entry)
        return None



from typing import Any, Iterator, List, Optional, Tuple

from .files import find_dandi_files

# TODO: provide our own "errors" records, which would also include warnings etc


def validate_bids(
    *paths: str,
    schema_version: Optional[str] = None,
    devel_debug: bool = False,
    report: Optional[str] = None,
) -> Any:
    """Validate BIDS paths.

    Parameters
    ----------
    paths : *str
        Paths to validate.
    schema_version : str, optional
        BIDS schema version to use, this setting will override the version specified in the dataset.
    devel_debug : bool, optional
        Whether to trigger debugging in the BIDS validator.
    report_path : bool or str, optional
        If `True` a log will be written using the standard output path of `.write_report()`.
        If string, the string will be used as the output path.
        If the variable evaluates as False, no log will be written.

    Notes
    -----
    Can be used from bash, as:
        DANDI_DEVEL=1 dandi validate-bids --schema="1.7.0+012+dandi001" --report="my.log" /my/path
    """
    from .bids_validator_xs import validate_bids as validate_bids_

    return validate_bids_(
        paths, schema_version=schema_version, debug=devel_debug, report_path=report
    )


def validate(
    *paths: str,
    schema_version: Optional[str] = None,
    devel_debug: bool = False,
    allow_any_path: bool = False,
) -> Iterator[Tuple[str, List[str]]]:
    """Validate content

    Parameters
    ----------
    paths: *str
      Could be individual (.nwb) files or a single dandiset path.

    Yields
    ------
    path, errors
      errors for a path
    """
    for df in find_dandi_files(*paths, dandiset_path=None, allow_all=allow_any_path):
        yield (
            str(df.filepath),
            df.get_validation_errors(
                schema_version=schema_version, devel_debug=devel_debug
            ),
        )



import re

import pytest
import responses

from dandi.consts import known_instances
from dandi.dandiarchive import (
    AssetFolderURL,
    AssetIDURL,
    AssetItemURL,
    AssetPathPrefixURL,
    BaseAssetIDURL,
    DandisetURL,
    ParsedDandiURL,
    follow_redirect,
    parse_dandi_url,
)
from dandi.exceptions import NotFoundError, UnknownURLError
from dandi.tests.skip import mark

from .fixtures import DandiAPI, SampleDandiset


@pytest.mark.parametrize(
    "url,parsed_url",
    [
        # New DANDI web UI driven by DANDI API.
        (
            "https://gui.dandiarchive.org/#/dandiset/000001",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000001",
                version_id=None,
            ),
        ),
        (
            "https://gui.dandiarchive.org/#/dandiset/000001/",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000001",
                version_id=None,
            ),
        ),
        (
            "https://gui.dandiarchive.org/#/dandiset/000001/0.201104.2302",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000001",
                version_id="0.201104.2302",
            ),
        ),
        (
            "https://gui.dandiarchive.org/#/dandiset/000001/0.201104.2302/",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000001",
                version_id="0.201104.2302",
            ),
        ),
        (
            "https://gui.dandiarchive.org/#/dandiset/000001/0.201104.2302/files",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000001",
                version_id="0.201104.2302",
            ),
        ),
        (
            "https://gui.dandiarchive.org/#/dandiset/000001/draft",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000001",
                version_id="draft",
            ),
        ),
        (
            "https://gui.dandiarchive.org/dandiset/000001",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000001",
                version_id=None,
            ),
        ),
        (
            "https://gui.dandiarchive.org/dandiset/000001/0.201104.2302",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000001",
                version_id="0.201104.2302",
            ),
        ),
        (
            "https://gui.dandiarchive.org/dandiset/000001/0.201104.2302/files",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000001",
                version_id="0.201104.2302",
            ),
        ),
        (
            "https://gui.dandiarchive.org/dandiset/000001/draft",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000001",
                version_id="draft",
            ),
        ),
        pytest.param(
            "DANDI:000027",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000027",
                version_id=None,
            ),
            marks=mark.skipif_no_network,
        ),
        pytest.param(
            "DANDI:000027/0.210831.2033",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000027",
                version_id="0.210831.2033",
            ),
            marks=mark.skipif_no_network,
        ),
        pytest.param(
            "dandi:000027/0.210831.2033",
            DandisetURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000027",
                version_id="0.210831.2033",
            ),
            marks=mark.skipif_no_network,
        ),
        (
            "http://localhost:8000/api/dandisets/000002/",
            DandisetURL(
                api_url="http://localhost:8000/api",
                dandiset_id="000002",
                version_id=None,
            ),
        ),
        (
            "http://localhost:8000/api/dandisets/000002",
            DandisetURL(
                api_url="http://localhost:8000/api",
                dandiset_id="000002",
                version_id=None,
            ),
        ),
        (
            "http://localhost:8000/api/dandisets/000002/versions/draft",
            DandisetURL(
                api_url="http://localhost:8000/api",
                dandiset_id="000002",
                version_id="draft",
            ),
        ),
        (
            "http://localhost:8000/api/dandisets/000002/versions/draft/",
            DandisetURL(
                api_url="http://localhost:8000/api",
                dandiset_id="000002",
                version_id="draft",
            ),
        ),
        (
            "https://gui.dandiarchive.org/#/dandiset/000001/files"
            "?location=%2Fsub-anm369962",
            AssetItemURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000001",
                version_id=None,
                path="sub-anm369962",
            ),
        ),
        (
            "https://gui.dandiarchive.org/#/dandiset/000006/0.200714.1807/files"
            "?location=%2Fsub-anm369962",
            AssetItemURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000006",
                version_id="0.200714.1807",
                path="sub-anm369962",
            ),
        ),
        (
            "https://gui.dandiarchive.org/#/dandiset/001001/draft/files"
            "?location=sub-RAT123%2F",
            AssetFolderURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="001001",
                version_id="draft",
                path="sub-RAT123/",
            ),
        ),
        # by direct instance name ad-hoc URI instance:ID[@version][/path]
        (
            "dandi://dandi-api-local-docker-tests/000002@draft",
            DandisetURL(
                api_url=known_instances["dandi-api-local-docker-tests"].api,
                dandiset_id="000002",
                version_id="draft",
            ),
        ),
        (
            "dandi://dandi-api-local-docker-tests/000002@draft/path",
            AssetItemURL(
                api_url=known_instances["dandi-api-local-docker-tests"].api,
                dandiset_id="000002",
                version_id="draft",
                path="path",
            ),
        ),
        (
            "dandi://dandi-api-local-docker-tests/000002/path",
            AssetItemURL(
                api_url=known_instances["dandi-api-local-docker-tests"].api,
                dandiset_id="000002",
                version_id=None,
                path="path",
            ),
        ),
        (  # test on "public" instance and have trailing / to signal the folder
            "dandi://dandi/000002/path/",
            AssetFolderURL(
                api_url=known_instances["dandi"].api,
                dandiset_id="000002",
                version_id=None,
                path="path/",
            ),
        ),
        (
            "https://api.dandiarchive.org/api/dandisets/000003/versions/draft"
            "/assets/0a748f90-d497-4a9c-822e-9c63811db412/download/",
            AssetIDURL(
                api_url="https://api.dandiarchive.org/api",
                dandiset_id="000003",
                version_id="draft",
                asset_id="0a748f90-d497-4a9c-822e-9c63811db412",
            ),
        ),
        (
            "https://api.dandiarchive.org/api/dandisets/000003/versions/draft"
            "/assets/0a748f90-d497-4a9c-822e-9c63811db412/download",
            AssetIDURL(
                api_url="https://api.dandiarchive.org/api",
                dandiset_id="000003",
                version_id="draft",
                asset_id="0a748f90-d497-4a9c-822e-9c63811db412",
            ),
        ),
        (
            "https://api.dandiarchive.org/api"
            "/assets/0a748f90-d497-4a9c-822e-9c63811db412/download/",
            BaseAssetIDURL(
                api_url="https://api.dandiarchive.org/api",
                asset_id="0a748f90-d497-4a9c-822e-9c63811db412",
            ),
        ),
        (
            "https://api.dandiarchive.org/api"
            "/assets/0a748f90-d497-4a9c-822e-9c63811db412/download",
            BaseAssetIDURL(
                api_url="https://api.dandiarchive.org/api",
                asset_id="0a748f90-d497-4a9c-822e-9c63811db412",
            ),
        ),
        (
            "https://api.dandiarchive.org/api/dandisets/000003/versions/draft"
            "/assets/?path=sub-YutaMouse20",
            AssetPathPrefixURL(
                api_url="https://api.dandiarchive.org/api",
                dandiset_id="000003",
                version_id="draft",
                path="sub-YutaMouse20",
            ),
        ),
        (
            "https://gui-staging.dandiarchive.org/#/dandiset/000018",
            DandisetURL(
                api_url=known_instances["dandi-staging"].api,
                dandiset_id="000018",
                version_id=None,
            ),
        ),
        (
            "https://deploy-preview-854--gui-dandiarchive-org.netlify.app"
            "/#/dandiset/000018",
            DandisetURL(
                api_url=known_instances["dandi-staging"].api,
                dandiset_id="000018",
                version_id=None,
            ),
        ),
    ],
)
def test_parse_api_url(url: str, parsed_url: ParsedDandiURL) -> None:
    assert parse_dandi_url(url) == parsed_url


@pytest.mark.parametrize(
    "url",
    [
        "DANDI:27",
        "DANDI:000027/draft",
        # Currently takes too long to run; cf. #830:
        # "https://identifiers.org/DANDI:000027/draft",
    ],
)
def test_parse_bad_api_url(url: str) -> None:
    with pytest.raises(UnknownURLError):
        parse_dandi_url(url)


def test_parse_dandi_url_unknown_instance() -> None:
    with pytest.raises(UnknownURLError) as excinfo:
        parse_dandi_url("dandi://not-an-instance/000001")
    assert str(excinfo.value) == (
        "Unknown instance 'not-an-instance'.  Valid instances: dandi,"
        " dandi-api-local-docker-tests, dandi-devel, dandi-staging"
    )


@mark.skipif_no_network
@pytest.mark.xfail(reason="https://github.com/dandi/dandi-archive/issues/1020")
def test_parse_dandi_url_not_found() -> None:
    # Unlikely this one would ever come to existence
    with pytest.raises(NotFoundError):
        parse_dandi_url("https://dandiarchive.org/dandiset/999999")


@mark.skipif_no_network
def test_follow_redirect() -> None:
    url = follow_redirect("https://bit.ly/dandi12")
    assert re.match(r"https://(.*\.)?dandiarchive.org", url)


@responses.activate
def test_parse_gui_new_redirect() -> None:
    redirector_base = known_instances["dandi"].redirector
    responses.add(
        responses.GET,
        f"{redirector_base}/server-info",
        json={
            "version": "1.2.0",
            "cli-minimal-version": "0.6.0",
            "cli-bad-versions": [],
            "services": {
                "webui": {"url": "https://gui.dandirchive.org"},
                "api": {"url": "https://api.dandiarchive.org/api"},
                "jupyterhub": {"url": "https://hub.dandiarchive.org"},
            },
        },
    )
    assert parse_dandi_url(
        "https://gui.dandiarchive.org/#/dandiset/000003"
    ) == DandisetURL(
        api_url="https://api.dandiarchive.org/api",
        dandiset_id="000003",
        version_id=None,
    )


@pytest.mark.parametrize("version_suffix", ["", "@draft", "@0.999999.9999"])
def test_get_nonexistent_dandiset(
    local_dandi_api: DandiAPI, version_suffix: str
) -> None:
    url = f"dandi://{local_dandi_api.instance_id}/999999{version_suffix}"
    parsed_url = parse_dandi_url(url)
    client = local_dandi_api.client
    parsed_url.get_dandiset(client)  # No error
    with pytest.raises(NotFoundError) as excinfo:
        parsed_url.get_dandiset(client, lazy=False)
    assert str(excinfo.value) == "No such Dandiset: '999999'"
    assert list(parsed_url.get_assets(client)) == []
    with pytest.raises(NotFoundError) as excinfo:
        next(parsed_url.get_assets(client, strict=True))
    assert str(excinfo.value) == "No such Dandiset: '999999'"


@pytest.mark.parametrize("version", ["draft", "0.999999.9999"])
def test_get_nonexistent_dandiset_asset_id(
    local_dandi_api: DandiAPI, version: str
) -> None:
    url = (
        f"{local_dandi_api.api_url}/dandisets/999999/versions/{version}"
        "/assets/00000000-0000-0000-0000-000000000000/"
    )
    parsed_url = parse_dandi_url(url)
    client = local_dandi_api.client
    assert list(parsed_url.get_assets(client)) == []
    with pytest.raises(NotFoundError) as excinfo:
        next(parsed_url.get_assets(client, strict=True))
    assert str(excinfo.value) == "No such Dandiset: '999999'"


def test_get_dandiset_nonexistent_asset_id(text_dandiset: SampleDandiset) -> None:
    url = (
        f"{text_dandiset.api.api_url}/dandisets/"
        f"{text_dandiset.dandiset_id}/versions/draft/assets/"
        "00000000-0000-0000-0000-000000000000/"
    )
    parsed_url = parse_dandi_url(url)
    client = text_dandiset.client
    assert list(parsed_url.get_assets(client)) == []
    with pytest.raises(NotFoundError) as excinfo:
        next(parsed_url.get_assets(client, strict=True))
    assert str(excinfo.value) == (
        "No such asset: '00000000-0000-0000-0000-000000000000' for"
        f" DANDI-API-LOCAL-DOCKER-TESTS:{text_dandiset.dandiset_id}/draft"
    )


def test_get_nonexistent_asset_id(local_dandi_api: DandiAPI) -> None:
    url = f"{local_dandi_api.api_url}/assets/00000000-0000-0000-0000-000000000000/"
    parsed_url = parse_dandi_url(url)
    client = local_dandi_api.client
    assert list(parsed_url.get_assets(client)) == []
    with pytest.raises(NotFoundError) as excinfo:
        next(parsed_url.get_assets(client, strict=True))
    assert str(excinfo.value) == "No such asset: '00000000-0000-0000-0000-000000000000'"


@pytest.mark.parametrize("version_suffix", ["", "@draft", "@0.999999.9999"])
def test_get_nonexistent_dandiset_asset_path(
    local_dandi_api: DandiAPI, version_suffix: str
) -> None:
    url = f"dandi://{local_dandi_api.instance_id}/999999{version_suffix}/does/not/exist"
    parsed_url = parse_dandi_url(url)
    client = local_dandi_api.client
    assert list(parsed_url.get_assets(client)) == []
    with pytest.raises(NotFoundError) as excinfo:
        next(parsed_url.get_assets(client, strict=True))
    assert str(excinfo.value) == "No such Dandiset: '999999'"


def test_get_nonexistent_asset_path(text_dandiset: SampleDandiset) -> None:
    url = (
        f"dandi://{text_dandiset.api.instance_id}/"
        f"{text_dandiset.dandiset_id}/does/not/exist"
    )
    parsed_url = parse_dandi_url(url)
    client = text_dandiset.client
    assert list(parsed_url.get_assets(client)) == []
    with pytest.raises(NotFoundError) as excinfo:
        next(parsed_url.get_assets(client, strict=True))
    assert str(excinfo.value) == "No asset at path 'does/not/exist'"


@pytest.mark.parametrize("version_suffix", ["", "@draft", "@0.999999.9999"])
def test_get_nonexistent_dandiset_asset_folder(
    local_dandi_api: DandiAPI, version_suffix: str
) -> None:
    url = (
        f"dandi://{local_dandi_api.instance_id}/999999{version_suffix}"
        "/does/not/exist/"
    )
    parsed_url = parse_dandi_url(url)
    client = local_dandi_api.client
    assert list(parsed_url.get_assets(client)) == []
    with pytest.raises(NotFoundError) as excinfo:
        next(parsed_url.get_assets(client, strict=True))
    assert str(excinfo.value) == "No such Dandiset: '999999'"


def test_get_nonexistent_asset_folder(text_dandiset: SampleDandiset) -> None:
    url = (
        f"dandi://{text_dandiset.api.instance_id}/"
        f"{text_dandiset.dandiset_id}/does/not/exist/"
    )
    parsed_url = parse_dandi_url(url)
    client = text_dandiset.client
    assert list(parsed_url.get_assets(client)) == []
    assert list(parsed_url.get_assets(client, strict=True)) == []


@pytest.mark.parametrize("version", ["draft", "0.999999.9999"])
def test_get_nonexistent_dandiset_asset_prefix(
    local_dandi_api: DandiAPI, version: str
) -> None:
    url = (
        f"{local_dandi_api.api_url}/dandisets/999999/versions/{version}"
        "/assets/?path=does/not/exist"
    )
    parsed_url = parse_dandi_url(url)
    client = local_dandi_api.client
    assert list(parsed_url.get_assets(client)) == []
    with pytest.raises(NotFoundError) as excinfo:
        next(parsed_url.get_assets(client, strict=True))
    assert str(excinfo.value) == "No such Dandiset: '999999'"


def test_get_nonexistent_asset_prefix(text_dandiset: SampleDandiset) -> None:
    url = (
        f"{text_dandiset.api.api_url}/dandisets/"
        f"{text_dandiset.dandiset_id}/versions/draft/assets/?path=does/not/exist"
    )
    parsed_url = parse_dandi_url(url)
    client = text_dandiset.client
    assert list(parsed_url.get_assets(client)) == []
    assert list(parsed_url.get_assets(client, strict=True)) == []



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the dandi package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from .. import digests
from ..digests import Digester, get_zarr_checksum


def test_digester(tmp_path):
    digester = Digester()

    f = tmp_path / "sample.txt"
    f.write_bytes(b"123")
    assert digester(f) == {
        "md5": "202cb962ac59075b964b07152d234b70",
        "sha1": "40bd001563085fc35165329ea1ff5c5ecbdbbeef",
        "sha256": "a665a45920422f9d417e4867efdc4fb8a04a1f3fff1fa07e998e86f7f7a27ae3",
        "sha512": "3c9909afec25354d551dae21590bb26e38d53f2173b8d3dc3eee4c047e7a"
        "b1c1eb8b85103e3be7ba613b31bb5c9c36214dc9f14a42fd7a2fdb84856b"
        "ca5c44c2",
    }

    f = tmp_path / "0"
    f.write_bytes(chr(0).encode())
    assert digester(f) == {
        "md5": "93b885adfe0da089cdf634904fd59f71",
        "sha1": "5ba93c9db0cff93f52b521d7420e43f6eda2784f",
        "sha256": "6e340b9cffb37a989ca544e6bb780a2c78901d3fb33738768511a30617afa01d",
        "sha512": "b8244d028981d693af7b456af8efa4cad63d282e19ff14942c246e50d935"
        "1d22704a802a71c3580b6370de4ceb293c324a8423342557d4e5c38438f0"
        "e36910ee",
    }

    f = tmp_path / "long.txt"
    f.write_bytes(b"123abz\n" * 1000000)
    assert digester(f) == {
        "md5": "81b196e3d8a1db4dd2e89faa39614396",
        "sha1": "5273ac6247322c3c7b4735a6d19fd4a5366e812f",
        "sha256": "80028815b3557e30d7cbef1d8dbc30af0ec0858eff34b960d2839fd88ad08871",
        "sha512": "684d23393eee455f44c13ab00d062980937a5d040259d69c6b291c983bf6"
        "35e1d405ff1dc2763e433d69b8f299b3f4da500663b813ce176a43e29ffc"
        "c31b0159",
    }


def test_get_zarr_checksum(mocker: MockerFixture, tmp_path: Path) -> None:
    # Use write_bytes() so that the line endings are the same on POSIX and
    # Windows.
    (tmp_path / "file1.txt").write_bytes(b"This is the first file.\n")
    (tmp_path / "file2.txt").write_bytes(b"This is the second file.\n")
    sub1 = tmp_path / "sub1"
    sub1.mkdir()
    (sub1 / "file3.txt").write_bytes(b"This is the third file.\n")
    (sub1 / "file4.txt").write_bytes(b"This is the fourth file.\n")
    (sub1 / "file5.txt").write_bytes(b"This is the fifth file.\n")
    subsub = sub1 / "subsub"
    subsub.mkdir()
    (subsub / "file6.txt").write_bytes(b"This is the sixth file.\n")
    sub2 = tmp_path / "sub2"
    sub2.mkdir()
    (sub2 / "file7.txt").write_bytes(b"This is the seventh file.\n")
    (sub2 / "file8.txt").write_bytes(b"This is the eighth file.\n")
    empty = tmp_path / "empty"
    empty.mkdir()

    assert (
        get_zarr_checksum(tmp_path / "file1.txt") == "d0aa42f003e36c1ecaf9aa8f20b6f1ad"
    )
    assert get_zarr_checksum(tmp_path) == "25627e0fc7c609d10100d020f7782a25-8--197"
    assert get_zarr_checksum(sub1) == "64af93ad7f8d471c00044d1ddbd4c0ba-4--97"

    with pytest.raises(ValueError) as excinfo:
        get_zarr_checksum(empty)
    assert str(excinfo.value) == "Cannot compute a Zarr checksum for an empty directory"

    spy = mocker.spy(digests, "md5file_nocache")
    assert (
        get_zarr_checksum(
            tmp_path,
            known={
                "file1.txt": "9ee7a152c5adb60803c928733acc1533",
                # ^^ This one is different!
                "file2.txt": "340c108ee69bf4626e7995a7048f52b8",
                "sub1/file3.txt": "7351dc767bfad322ddce50401badc359",
                "sub1/file4.txt": "bbede70f39fa8fc34f2dc4eda8b6bdea",
                "sub1/file5.txt": "c4e828c509f90b84e5b72d9d5612d676",
                "sub1/subsub/file6.txt": "6a7fe3b9e2c69a54216b7d5dcb4fe61d",
                # "sub2/file7.txt": Absent!
                "sub2/file8.txt": "7aadbff2b21f438baccded18b2e81ae3",
                "nonexistent-file.txt": "123456789012345678901234567890ab",
                # ^^ Not used in calculation!
            },
        )
        == "f77f4c5b277575f781c19ba91422f0c5-8--197"
    )
    spy.assert_called_once_with(sub2 / "file7.txt")



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the dandi package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Provides helper to compute digests (md5 etc) on files
"""

from __future__ import annotations

from dataclasses import dataclass, field
import hashlib
import logging
import os.path
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union, cast

from dandischema.digests.dandietag import DandiETag
from dandischema.digests.zarr import get_checksum
from fscacher import PersistentCache

from .threaded_walk import threaded_walk
from ..utils import auto_repr

lgr = logging.getLogger("dandi.support.digests")


@auto_repr
class Digester:
    """Helper to compute multiple digests in one pass for a file"""

    # Loosely based on snippet by PM 2Ring 2014.10.23
    # http://unix.stackexchange.com/a/163769/55543

    # Ideally we should find an efficient way to parallelize this but
    # atm this one is sufficiently speedy

    DEFAULT_DIGESTS = ["md5", "sha1", "sha256", "sha512"]

    def __init__(
        self, digests: Optional[List[str]] = None, blocksize: int = 1 << 16
    ) -> None:
        """
        Parameters
        ----------
        digests : list or None
          List of any supported algorithm labels, such as md5, sha1, etc.
          If None, a default set of hashes will be computed (md5, sha1,
          sha256, sha512).
        blocksize : int
          Chunk size (in bytes) by which to consume a file.
        """
        self._digests = digests or self.DEFAULT_DIGESTS
        self._digest_funcs = [getattr(hashlib, digest) for digest in self._digests]
        self.blocksize = blocksize

    @property
    def digests(self) -> List[str]:
        return self._digests

    def __call__(self, fpath: Union[str, Path]) -> Dict[str, str]:
        """
        fpath : str
          File path for which a checksum shall be computed.

        Return
        ------
        dict
          Keys are algorithm labels, and values are checksum strings
        """
        lgr.debug("Estimating digests for %s" % fpath)
        digests = [x() for x in self._digest_funcs]
        with open(fpath, "rb") as f:
            while True:
                block = f.read(self.blocksize)
                if not block:
                    break
                [d.update(block) for d in digests]

        return {n: d.hexdigest() for n, d in zip(self.digests, digests)}


checksums = PersistentCache(name="dandi-checksums", envvar="DANDI_CACHE")


@checksums.memoize_path
def get_digest(filepath: Union[str, Path], digest: str = "sha256") -> str:
    if digest == "dandi-etag":
        return cast(str, get_dandietag(filepath).as_str())
    elif digest == "zarr-checksum":
        return get_zarr_checksum(Path(filepath))
    else:
        return Digester([digest])(filepath)[digest]


@checksums.memoize_path
def get_dandietag(filepath: Union[str, Path]) -> DandiETag:
    return DandiETag.from_file(filepath)


def get_zarr_checksum(path: Path, known: Optional[Dict[str, str]] = None) -> str:
    """
    Compute the Zarr checksum for a file or directory tree.

    If the digests for any files in the Zarr are already known, they can be
    passed in the ``known`` argument, which must be a `dict` mapping
    slash-separated paths relative to the root of the Zarr to hex digests.
    """
    if path.is_file():
        return cast(str, get_digest(path, "md5"))
    if known is None:
        known = {}

    def digest_file(f: Path) -> Tuple[Path, str, int]:
        assert known is not None
        relpath = f.relative_to(path).as_posix()
        try:
            dgst = known[relpath]
        except KeyError:
            dgst = md5file_nocache(f)
        return (f, dgst, os.path.getsize(f))

    zcc = ZCDirectory()
    for p, digest, size in threaded_walk(path, digest_file):
        zcc.add(p.relative_to(path), digest, size)
    return zcc.get_digest_size()[0]


@dataclass
class ZCFile:
    """
    File node used for building an in-memory tree of Zarr entries and their
    digests when calculating a complete Zarr checksum

    :meta private:
    """

    digest: str
    size: int

    def get_digest_size(self) -> Tuple[str, int]:
        return (self.digest, self.size)


@dataclass
class ZCDirectory:
    """
    Directory node used for building an in-memory tree of Zarr entries and
    their digests when calculating a complete Zarr checksum

    :meta private:
    """

    children: Dict[str, Union[ZCDirectory, ZCFile]] = field(default_factory=dict)

    def get_digest_size(self) -> Tuple[str, int]:
        size = 0
        files = {}
        dirs = {}
        for name, n in self.children.items():
            dgst, sz = n.get_digest_size()
            if isinstance(n, ZCDirectory):
                dirs[name] = (dgst, sz)
            else:
                files[name] = (dgst, sz)
            size += sz
        return (cast(str, get_checksum(files, dirs)), size)

    def add(self, path: Path, digest: str, size: int) -> None:
        *dirs, name = path.parts
        parts = []
        d = self
        for dirname in dirs:
            parts.append(dirname)
            e = d.children.setdefault(dirname, ZCDirectory())
            assert isinstance(
                e, ZCDirectory
            ), f"Path type conflict for {'/'.join(parts)}"
            d = e
        parts.append(name)
        pstr = "/".join(parts)
        assert name not in d.children, f"File {pstr} encountered twice"
        d.children[name] = ZCFile(digest=digest, size=size)


def md5file_nocache(filepath: Union[str, Path]) -> str:
    """
    Compute the MD5 digest of a file without caching with fscacher, which has
    been shown to slow things down for the large numbers of files typically
    present in Zarrs
    """
    return Digester(["md5"])(filepath)["md5"]



"""
THIS MODULE IS BUNDLED FROM THE bids-specification PACKAGE.
Schema loading- and processing-related functions.
"""
from copy import deepcopy
import logging
import os
from pathlib import Path

from ruamel import yaml

from . import utils

lgr = utils.get_logger()
# Basic settings for output, for now just basic
utils.set_logger_level(lgr, os.environ.get("BIDS_SCHEMA_LOG_LEVEL", logging.INFO))
logging.basicConfig(format="%(asctime)-15s [%(levelname)8s] %(message)s")


def _get_entry_name(path):
    if path.suffix == ".yaml":
        return path.name[:-5]  # no .yaml
    else:
        return path.name


def dereference_yaml(schema, struct):
    """Recursively search a dictionary-like object for $ref keys.

    Each $ref key is replaced with the contents of the referenced field in the overall
    dictionary-like object.
    """
    if isinstance(struct, dict):
        if "$ref" in struct:
            ref_field = struct["$ref"]
            template = schema[ref_field]
            struct.pop("$ref")
            # Result is template object with local overrides
            struct = {**template, **struct}

        struct = {key: dereference_yaml(schema, val) for key, val in struct.items()}

    elif isinstance(struct, list):
        struct = [dereference_yaml(schema, item) for item in struct]

    return struct


def load_schema(schema_path):
    """Load the schema into a dictionary.

    This function allows the schema, like BIDS itself, to be specified in
    a hierarchy of directories and files.
    File names (minus extensions) and directory names become keys
    in the associative array (dict) of entries composed from content
    of files and entire directories.

    Parameters
    ----------
    schema_path : str
        Folder containing yaml files or yaml file.

    Returns
    -------
    dict
        Schema in dictionary form.
    """
    _yaml = yaml.YAML(typ="safe", pure=True)

    schema_path = Path(schema_path)
    objects_dir = schema_path / "objects/"
    rules_dir = schema_path / "rules/"

    if not objects_dir.is_dir() or not rules_dir.is_dir():
        raise ValueError(
            f"Schema path or paths do not exist:\n\t{str(objects_dir)}\n\t{str(rules_dir)}"
        )

    schema = {}
    schema["objects"] = {}
    schema["rules"] = {}

    # Load object definitions. All are present in single files.
    for object_group_file in sorted(objects_dir.glob("*.yaml")):
        lgr.debug(f"Loading {object_group_file.stem} objects.")
        dict_ = _yaml.load(object_group_file.read_text())
        schema["objects"][object_group_file.stem] = dereference_yaml(dict_, dict_)

    # Grab single-file rule groups
    for rule_group_file in sorted(rules_dir.glob("*.yaml")):
        lgr.debug(f"Loading {rule_group_file.stem} rules.")
        dict_ = _yaml.load(rule_group_file.read_text())
        schema["rules"][rule_group_file.stem] = dereference_yaml(dict_, dict_)

    # Load folders of rule subgroups.
    for rule_group_file in sorted(rules_dir.glob("*/*.yaml")):
        rule = schema["rules"].setdefault(rule_group_file.parent.name, {})
        lgr.debug(f"Loading {rule_group_file.stem} rules.")
        dict_ = _yaml.load(rule_group_file.read_text())
        rule[rule_group_file.stem] = dereference_yaml(dict_, dict_)

    return schema


def filter_schema(schema, **kwargs):
    """Filter the schema based on a set of keyword arguments.

    Parameters
    ----------
    schema : dict
        The schema object, which is a dictionary with nested dictionaries and
        lists stored within it.
    kwargs : dict
        Keyword arguments used to filter the schema.
        Example kwargs that may be used include: "suffixes", "datatypes",
        "extensions".

    Returns
    -------
    new_schema : dict
        The filtered version of the schema.

    Notes
    -----
    This function calls itself recursively, in order to apply filters at
    arbitrary depth.

    Warning
    -------
    This function employs a *very* simple filter. It is very limited.
    """
    new_schema = deepcopy(schema)
    if isinstance(new_schema, dict):
        # Reduce values in dict to only requested
        for k, v in kwargs.items():
            if k in new_schema.keys():
                filtered_item = deepcopy(new_schema[k])
                if isinstance(filtered_item, dict):
                    filtered_item = {
                        k1: v1 for k1, v1 in filtered_item.items() if k1 in v
                    }
                else:
                    filtered_item = [i for i in filtered_item if i in v]
                new_schema[k] = filtered_item

            for k2, v2 in new_schema.items():
                new_schema[k2] = filter_schema(new_schema[k2], **kwargs)

    elif isinstance(new_schema, list):
        for i, item in enumerate(new_schema):
            if isinstance(item, dict):
                new_schema[i] = filter_schema(item, **kwargs)
    return new_schema



import os

BIDS_EXAMPLES_BLACKLIST = [
    "invalid_pet001",
]
BIDS_EXAMPLES_WHITELIST = [
    "asl003",
    "eeg_cbm",
    "hcp_example_bids",
    "micr_SEM",
    "micr_SEM-dandi",
    "micr_SPIM",
    "pet001",
    "pet003",
    "qmri_megre",
    "qmri_tb1tfl",
    "qmri_vfa",
]
TEST_SCHEMA_PATH = "{module_path}/support/bids/schemadata/1.7.0+012+dandi001"


def test__add_entity():
    from dandi.bids_validator_xs import _add_entity

    # Test empty input and directory creation and required entity
    regex_entities = ""
    entity = "subject"
    entity_shorthand = "sub"
    variable_field = "([a-z,A-Z,0-9]*?)"
    requirement_level = "required"

    _regex_entities = _add_entity(
        regex_entities,
        entity,
        entity_shorthand,
        variable_field,
        requirement_level,
    )

    assert _regex_entities == "sub-(?P=subject)"

    # Test append input and optional entity
    regex_entities = (
        "sub-(?P=subject)(|_ses-(?P=session))"
        "(|_task-(?P<task>([a-z,A-Z,0-9]*?)))(|_trc-(?P<tracer>([a-z,A-Z,0-9]*?)))"
        "(|_rec-(?P<reconstruction>([a-z,A-Z,0-9]*?)))"
        "(|_run-(?P<run>([a-z,A-Z,0-9]*?)))"
    )
    entity = "recording"
    entity_shorthand = "recording"
    variable_field = "([a-z,A-Z,0-9]*?)"
    requirement_level = "optional"

    _regex_entities = _add_entity(
        regex_entities,
        entity,
        entity_shorthand,
        variable_field,
        requirement_level,
    )

    assert (
        _regex_entities == "sub-(?P=subject)(|_ses-(?P=session))"
        "(|_task-(?P<task>([a-z,A-Z,0-9]*?)))(|_trc-(?P<tracer>([a-z,A-Z,0-9]*?)))"
        "(|_rec-(?P<reconstruction>([a-z,A-Z,0-9]*?)))"
        "(|_run-(?P<run>([a-z,A-Z,0-9]*?)))"
        "(|_recording-(?P<recording>([a-z,A-Z,0-9]*?)))"
    )


def test__add_extensions():
    from dandi.bids_validator_xs import _add_extensions

    # Test single extension
    regex_string = (
        "sub-(?P=subject)(|_ses-(?P=session))"
        "_sample-(?P<sample>([a-z,A-Z,0-9]*?))"
        "(|_acq-(?P<acquisition>([a-z,A-Z,0-9]*?)))_photo"
    )
    variant = {
        "suffixes": ["photo"],
        "extensions": [".jpg"],
        "entities": {
            "subject": "required",
            "session": "optional",
            "sample": "required",
            "acquisition": "optional",
        },
    }
    _regex_string = _add_extensions(regex_string, variant)

    assert (
        _regex_string == "sub-(?P=subject)(|_ses-(?P=session))"
        "_sample-(?P<sample>([a-z,A-Z,0-9]*?))"
        "(|_acq-(?P<acquisition>([a-z,A-Z,0-9]*?)))_photo\\.jpg"
    )

    # Test multiple extensions
    regex_string = (
        "sub-(?P=subject)(|_ses-(?P=session))"
        "_sample-(?P<sample>([a-z,A-Z,0-9]*?))"
        "(|_acq-(?P<acquisition>([a-z,A-Z,0-9]*?)))_photo"
    )
    variant = {
        "suffixes": ["photo"],
        "extensions": [".jpg", ".png", ".tif"],
        "entities": {
            "subject": "required",
            "session": "optional",
            "sample": "required",
            "acquisition": "optional",
        },
    }
    _regex_string = _add_extensions(regex_string, variant)

    assert (
        _regex_string == "sub-(?P=subject)(|_ses-(?P=session))"
        "_sample-(?P<sample>([a-z,A-Z,0-9]*?))"
        "(|_acq-(?P<acquisition>([a-z,A-Z,0-9]*?)))"
        "_photo(\\.jpg|\\.png|\\.tif)"
    )


def test__add_subdirs():
    from dandi.bids_validator_xs import _add_subdirs

    regex_string = "sub-(?P=subject)_sessions\\.(tsv|json)"
    variant = {
        "suffixes": ["sessions"],
        "extensions": [".tsv", ".json"],
        "entities": {"subject": "required"},
    }
    datatype = "tabular_metadata"
    entity_definitions = {
        "acquisition": {
            "name": "Acquisition",
            "entity": "acq",
            "type": "string",
            "format": "label",
        },
        "session": {
            "name": "Session",
            "entity": "ses",
            "type": "string",
            "format": "label",
        },
        "subject": {
            "name": "Subject",
            "entity": "sub",
            "type": "string",
            "format": "label",
        },
    }
    modality_datatypes = [
        "anat",
        "dwi",
        "fmap",
        "func",
        "perf",
        "eeg",
        "ieeg",
        "meg",
        "beh",
        "pet",
        "micr",
    ]
    _regex_string = _add_subdirs(
        regex_string, variant, datatype, entity_definitions, modality_datatypes
    )

    assert (
        _regex_string == "/sub-(?P<subject>([a-z,A-Z,0-9]*?))/sub-(?P=subject)"
        "_sessions\\.(tsv|json)"
    )


def test__add_suffixes():
    from dandi.bids_validator_xs import _add_suffixes

    # Test single expansion
    regex_entities = "sub-(?P=subject)"
    variant = {
        "suffixes": ["sessions"],
        "extensions": [
            ".tsv",
            ".json",
        ],
        "entities": {"subject": "required"},
    }
    regex_string = "sub-(?P=subject)_sessions"

    _regex_string = _add_suffixes(regex_entities, variant)

    assert _regex_string == regex_string

    # Test multiple expansions
    regex_entities = (
        "sub-(?P=subject)(|_ses-(?P=session))"
        "(|_acq-(?P<acquisition>([a-z,A-Z,0-9]*?)))"
        "(|_rec-(?P<reconstruction>([a-z,A-Z,0-9]*?)))"
        "(|_dir-(?P<direction>([a-z,A-Z,0-9]*?)))(|_run-(?P<run>([a-z,A-Z,0-9]*?)))"
        "(|_recording-(?P<recording>([a-z,A-Z,0-9]*?)))"
    )
    variant = {
        "suffixes": [
            "physio",
            "stim",
        ],
        "extensions": [
            ".tsv.gz",
            ".json",
        ],
        "entities": {
            "subject": "required",
            "session": "optional",
            "acquisition": "optional",
            "reconstruction": "optional",
            "direction": "optional",
            "run": "optional",
            "recording": "optional",
        },
    }
    regex_string = (
        "sub-(?P=subject)(|_ses-(?P=session))"
        "(|_acq-(?P<acquisition>([a-z,A-Z,0-9]*?)))"
        "(|_rec-(?P<reconstruction>([a-z,A-Z,0-9]*?)))"
        "(|_dir-(?P<direction>([a-z,A-Z,0-9]*?)))(|_run-(?P<run>([a-z,A-Z,0-9]*?)))"
        "(|_recording-(?P<recording>([a-z,A-Z,0-9]*?)))"
        "_(physio|stim)"
    )

    _regex_string = _add_suffixes(regex_entities, variant)

    assert _regex_string == regex_string


def test_load_all():
    from dandi.bids_validator_xs import load_all

    schema_path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        "../support/bids/schemadata/1.7.0+012+dandi001",
    )
    schema_all = load_all(schema_path)

    # Check if expected keys are present in all entries
    for entry in schema_all:
        assert "regex" in list(entry.keys())
        assert "mandatory" in list(entry.keys())


def test_write_report(tmp_path):
    from dandi.bids_validator_xs import write_report

    validation_result = {}

    validation_result["schema_tracking"] = [
        {
            "regex": ".*?/sub-(?P<subject>([a-z,A-Z,0-9]*?))/"
            "(|ses-(?P<session>([a-z,A-Z,0-9]*?))/)anat/sub-(?P=subject)"
            "(|_ses-(?P=session))(|_acq-(?P<acquisition>([a-z,A-Z,0-9]*?)))"
            "(|_ce-(?P<ceagent>([a-z,A-Z,0-9]*?)))"
            "(|_rec-(?P<reconstruction>([a-z,A-Z,0-9]*?)))"
            "(|_run-(?P<run>([a-z,A-Z,0-9]*?)))"
            "(|_part-(?P<part>(mag|phase|real|imag)))"
            "_(T1w|T2w|PDw|T2starw|FLAIR|inplaneT1|inplaneT2|PDT2|angio|T2star)"
            "\\.(nii.gz|nii|json)$",
            "mandatory": False,
        }
    ]
    validation_result["schema_listing"] = [
        {
            "regex": ".*?/sub-(?P<subject>([a-z,A-Z,0-9]*?))/"
            "(|ses-(?P<session>([a-z,A-Z,0-9]*?))/)anat/sub-(?P=subject)"
            "(|_ses-(?P=session))(|_acq-(?P<acquisition>([a-z,A-Z,0-9]*?)))"
            "(|_ce-(?P<ceagent>([a-z,A-Z,0-9]*?)))"
            "(|_rec-(?P<reconstruction>([a-z,A-Z,0-9]*?)))"
            "(|_run-(?P<run>([a-z,A-Z,0-9]*?)))"
            "(|_part-(?P<part>(mag|phase|real|imag)))"
            "_(T1w|T2w|PDw|T2starw|FLAIR|inplaneT1|inplaneT2|PDT2|angio|T2star)"
            "\\.(nii.gz|nii|json)$",
            "mandatory": False,
        }
    ]
    validation_result["path_tracking"] = [
        "/home/chymera/.data2/datalad/000026/"
        "rawdata/sub-EXC022/anat/sub-EXC022_ses-MRI_flip-1_VFA.nii.gz"
    ]
    validation_result["path_listing"] = [
        "/home/chymera/.data2/datalad/000026/"
        "rawdata/sub-EXC022/anat/sub-EXC022_ses-MRI_flip-1_VFA.nii.gz"
    ]

    report_path = os.path.join(
        tmp_path,
        "output_bids_validator_xs_write.log",
    )
    expected_report_path = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        "data/expected_bids_validator_xs_write.log",
    )
    write_report(validation_result, report_path=report_path)
    with open(report_path, "r") as f:
        report_text = f.read()
    with open(expected_report_path, "r") as f:
        expected_report_text = f.read()
    assert report_text == expected_report_text


def test_bids_datasets(bids_examples):
    from dandi.bids_validator_xs import validate_bids

    # Validate per dataset, with automatic schema selection:
    for i in os.listdir(bids_examples):
        if i in BIDS_EXAMPLES_WHITELIST:
            result = validate_bids(
                os.path.join(bids_examples, i),
            )
            # Have all files been validated?
            assert len(result["path_tracking"]) == 0


def test_error_datasets(bids_examples):
    from dandi.bids_validator_xs import validate_bids

    # Validate per dataset, with automatic schema selection:
    for i in os.listdir(bids_examples):
        if i in BIDS_EXAMPLES_BLACKLIST:
            result = validate_bids(
                os.path.join(bids_examples, i),
            )
            # Are there non-validated files?
            assert len(result["path_tracking"]) != 0


def test_bids_datasets_selected_paths(bids_examples, tmp_path):
    from dandi.bids_validator_xs import validate_bids

    # Create input for file list based validation
    selected_dir = os.path.join(bids_examples, BIDS_EXAMPLES_WHITELIST[0])
    selected_paths = []
    for root, dirs, files in os.walk(selected_dir, topdown=False):
        for f in files:
            selected_path = os.path.join(root, f)
            selected_paths.append(selected_path)
    # Does explicit schema specification work?
    result = validate_bids(selected_paths, schema_version=TEST_SCHEMA_PATH)

    # Does terminal debug output work?
    result = validate_bids(selected_paths, debug=True)

    # Does the default report path work?
    result = validate_bids(selected_paths, report_path=True)

    # Does custom report path specification work?
    result = validate_bids(
        selected_paths,
        schema_version=TEST_SCHEMA_PATH,
        report_path=os.path.join(tmp_path, "test_bids.log"),
    )
    # Have all files been validated?
    assert len(result["path_tracking"]) == 0



import os

from click.testing import CliRunner

from ..command import instances


def test_cmd_instances(monkeypatch):
    redirector_base = os.environ.get(
        "DANDI_REDIRECTOR_BASE", "https://dandiarchive.org"
    )
    instancehost = os.environ.get("DANDI_INSTANCEHOST", "localhost")
    r = CliRunner().invoke(instances, [])
    assert r.exit_code == 0
    assert r.output == (
        "dandi:\n"
        "  api: https://api.dandiarchive.org/api\n"
        "  gui: https://gui.dandiarchive.org\n"
        f"  redirector: {redirector_base}\n"
        "dandi-api-local-docker-tests:\n"
        f"  api: http://{instancehost}:8000/api\n"
        "  gui: null\n"
        "  redirector: null\n"
        "dandi-devel:\n"
        "  api: null\n"
        "  gui: https://gui-beta-dandiarchive-org.netlify.app\n"
        "  redirector: null\n"
        "dandi-staging:\n"
        "  api: https://api-staging.dandiarchive.org/api\n"
        "  gui: https://gui-staging.dandiarchive.org\n"
        "  redirector: null\n"
    )
